#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.optim as optim
import torch.nn.functional as F
import parl
import numpy as np
from copy import deepcopy

__all__ = ['QMIX']


class QMIX(parl.Algorithm):
    def __init__(self,
                 agent_model,
                 qmixer_model,
                 double_q=True,
                 gamma=0.99,
                 lr=0.0005,
                 clip_grad_norm=None):
        """ QMIX algorithm
        Args:
            agent_model (parl.Model): agents' local q network for decision making.
            qmixer_model (parl.Model): A mixing network which takes local q values as input
                to construct a global Q network.
            double_q (bool): Double-DQN.
            gamma (float): discounted factor for reward computation.
            lr (float): learning rate.
            clip_grad_norm (None, or float): clipped value of gradients' global norm.
        """
        self.agent_model = agent_model
        self.qmixer_model = qmixer_model
        self.target_agent_model = deepcopy(self.agent_model)
        self.target_qmixer_model = deepcopy(self.qmixer_model)

        self.n_agents = self.qmixer_model.n_agents
        assert isinstance(gamma, float)
        assert isinstance(lr, float)
        self.double_q = double_q
        self.gamma = gamma
        self.lr = lr
        self.clip_grad_norm = clip_grad_norm

        device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        self.agent_model.to(device)
        self.target_agent_model.to(device)
        self.qmixer_model.to(device)
        self.target_qmixer_model.to(device)

        self.params = list(self.agent_model.parameters())
        self.params += self.qmixer_model.parameters()
        self.optimizer = torch.optim.RMSprop(
            params=self.params, lr=self.lr, alpha=0.99, eps=0.00001)

    def _init_hidden_states(self, batch_size):
        self.hidden_states = self.agent_model.init_hidden().unsqueeze(
            0).expand(batch_size, self.n_agents, -1)
        self.target_hidden_states = self.target_agent_model.init_hidden(
        ).unsqueeze(0).expand(batch_size, self.n_agents, -1)

    def predict_local_q(self, obs, hidden_state):
        '''
        Args:
            obs (torch.Tensor): (n_agents, obs_shape)
        Returns:
            self.agent_model(obs, hidden_state)
        '''
        return self.agent_model(obs, hidden_state)

    def learn(self, state_batch, actions_batch, reward_batch, terminated_batch,
              obs_batch, available_actions_batch, filled_batch):
        '''
        Args:
            state_batch (torch.Tensor):             (batch_size, T, state_shape)
            actions_batch (torch.Tensor):           (batch_size, T, n_agents)
            reward_batch (torch.Tensor):            (batch_size, T, 1)
            terminated_batch (torch.Tensor):        (batch_size, T, 1)
            obs_batch (torch.Tensor):               (batch_size, T, n_agents, obs_shape)
            available_actions_batch (torch.Tensor): (batch_size, T, n_agents, n_actions)
            filled_batch (torch.Tensor):            (batch_size, T, 1)
        Returns:
            loss (float): train loss
            td_error (float): train TD error
        '''
        batch_size = state_batch.shape[0]
        episode_len = state_batch.shape[1]
        self._init_hidden_states(batch_size)

        reward_batch = reward_batch[:, :-1, :]
        actions_batch = actions_batch[:, :-1, :].unsqueeze(-1)
        terminated_batch = terminated_batch[:, :-1, :]
        filled_batch = filled_batch[:, :-1, :]

        mask = (1 - filled_batch) * (1 - terminated_batch)

        local_qs = []
        target_local_qs = []
        for t in range(episode_len):
            obs = obs_batch[:, t, :, :]
            obs = obs.reshape(-1, obs_batch.shape[-1])
            local_q, self.hidden_states = self.agent_model(
                obs, self.hidden_states)
            local_q = local_q.reshape(batch_size, self.n_agents, -1)
            local_qs.append(local_q)

            target_local_q, self.target_hidden_states = self.target_agent_model(
                obs, self.target_hidden_states)
            target_local_q = target_local_q.view(batch_size, self.n_agents, -1)
            target_local_qs.append(target_local_q)

        local_qs = torch.stack(local_qs, dim=1)
        target_local_qs = torch.stack(target_local_qs[1:], dim=1)

        chosen_action_local_qs = torch.gather(
            local_qs[:, :-1, :, :], dim=3, index=actions_batch).squeeze(3)
        # mask unavailable actions
        target_local_qs[available_actions_batch[:, 1:, :] == 0] = -1e10
        if self.double_q:
            local_qs_detach = local_qs.clone().detach()
            local_qs_detach[available_actions_batch == 0] = -1e10
            cur_max_actions = local_qs_detach[:, 1:].max(
                dim=3, keepdim=True)[1]
            target_local_max_qs = torch.gather(target_local_qs, 3,
                                               cur_max_actions).squeeze(3)
        else:
            target_local_max_qs = target_local_qs.max(
                dim=3)[0]  # idx0: value, idx1: index

        chosen_action_global_qs = self.qmixer_model(chosen_action_local_qs,
                                                    state_batch[:, :-1, :])
        target_global_max_qs = self.target_qmixer_model(
            target_local_max_qs, state_batch[:, 1:, :])

        target = reward_batch + self.gamma * (
            1 - terminated_batch) * target_global_max_qs
        td_error = target.detach() - chosen_action_global_qs
        masked_td_error = td_error * mask
        mean_td_error = masked_td_error.sum() / mask.sum()
        loss = (masked_td_error**2).sum() / mask.sum()

        self.optimizer.zero_grad()
        loss.backward()
        if self.clip_grad_norm:
            torch.nn.utils.clip_grad_norm_(self.params, self.clip_grad_norm)
        self.optimizer.step()
        return loss.item(), mean_td_error.item()



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file wraps Fluid layers that have parameters to support parameter sharing.
For other layers that don't have parameters, we simply copy them to this namespace.
"""
from paddle.fluid.layers import *
from parl.core.fluid.layers.layer_wrappers import *



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
warnings.simplefilter('default')

import numpy as np
from copy import deepcopy
from paddle import fluid
from parl.core.fluid import layers
from parl.core.fluid.algorithm import Algorithm

__all__ = ['PPO']


class PPO(Algorithm):
    def __init__(self,
                 model,
                 act_dim=None,
                 policy_lr=None,
                 value_lr=None,
                 epsilon=0.2):
        """ PPO algorithm
        
        Args:
            model (parl.Model): model defining forward network of policy and value.
            act_dim (float): dimension of the action space.
            policy_lr (float): learning rate of the policy model. 
            value_lr (float): learning rate of the value model.
            epsilon (float): epsilon used in the CLIP loss (default 0.2).
        """
        self.model = model
        # Used to calculate probability of action in old policy
        self.old_policy_model = deepcopy(model.policy_model)

        assert isinstance(act_dim, int)
        assert isinstance(policy_lr, float)
        assert isinstance(value_lr, float)
        assert isinstance(epsilon, float)
        self.act_dim = act_dim
        self.policy_lr = policy_lr
        self.value_lr = value_lr
        self.epsilon = epsilon

    def _calc_logprob(self, actions, means, logvars):
        """ Calculate log probabilities of actions, when given means and logvars
            of normal distribution.
            The constant sqrt(2 * pi) is omitted, which will be eliminated in later.

        Args:
            actions: shape (batch_size, act_dim)
            means:   shape (batch_size, act_dim)
            logvars: shape (act_dim)

        Returns:
            logprob: shape (batch_size)
        """
        exp_item = layers.elementwise_div(
            layers.square(actions - means), layers.exp(logvars), axis=1)
        exp_item = -0.5 * layers.reduce_sum(exp_item, dim=1)

        vars_item = -0.5 * layers.reduce_sum(logvars)
        logprob = exp_item + vars_item
        return logprob

    def _calc_kl(self, means, logvars, old_means, old_logvars):
        """ Calculate KL divergence between old and new distributions
            See: https://en.wikipedia.org/wiki/Multivariate_normal_distribution#Kullback.E2.80.93Leibler_divergence

        Args:
            means: shape (batch_size, act_dim)
            logvars: shape (act_dim)
            old_means: shape (batch_size, act_dim)
            old_logvars: shape (act_dim)

        Returns:
            kl: shape (batch_size)
        """
        log_det_cov_old = layers.reduce_sum(old_logvars)
        log_det_cov_new = layers.reduce_sum(logvars)
        tr_old_new = layers.reduce_sum(layers.exp(old_logvars - logvars))
        kl = 0.5 * (layers.reduce_sum(
            layers.square(means - old_means) / layers.exp(logvars), dim=1) + (
                log_det_cov_new - log_det_cov_old) + tr_old_new - self.act_dim)
        return kl

    def predict(self, obs):
        """ Use the policy model of self.model to predict means and logvars of actions
        """
        means, logvars = self.model.policy(obs)
        return means

    def sample(self, obs):
        """ Use the policy model of self.model to sample actions
        """
        sampled_act = self.model.policy_sample(obs)
        return sampled_act

    def policy_learn(self, obs, actions, advantages, beta=None):
        """ Learn policy model with: 
                1. CLIP loss: Clipped Surrogate Objective 
                2. KLPEN loss: Adaptive KL Penalty Objective
            See: https://arxiv.org/pdf/1707.02286.pdf

        Args:
            obs: Tensor, (batch_size, obs_dim)
            actions: Tensor, (batch_size, act_dim)
            advantages: Tensor (batch_size, )
            beta: Tensor (1) or None
                  if None, use CLIP Loss; else, use KLPEN loss. 
        """
        old_means, old_logvars = self.old_policy_model.policy(obs)
        old_means.stop_gradient = True
        old_logvars.stop_gradient = True
        old_logprob = self._calc_logprob(actions, old_means, old_logvars)

        means, logvars = self.model.policy(obs)
        logprob = self._calc_logprob(actions, means, logvars)

        kl = self._calc_kl(means, logvars, old_means, old_logvars)
        kl = layers.reduce_mean(kl)

        if beta is None:  # Clipped Surrogate Objective
            pg_ratio = layers.exp(logprob - old_logprob)
            clipped_pg_ratio = layers.clip(pg_ratio, 1 - self.epsilon,
                                           1 + self.epsilon)
            surrogate_loss = layers.elementwise_min(
                advantages * pg_ratio, advantages * clipped_pg_ratio)
            loss = 0 - layers.reduce_mean(surrogate_loss)
        else:  # Adaptive KL Penalty Objective
            # policy gradient loss
            loss1 = 0 - layers.reduce_mean(
                advantages * layers.exp(logprob - old_logprob))
            # adaptive kl loss
            loss2 = kl * beta
            loss = loss1 + loss2
        optimizer = fluid.optimizer.AdamOptimizer(self.policy_lr)
        optimizer.minimize(loss)
        return loss, kl

    def value_predict(self, obs):
        """ Use value model of self.model to predict value of obs
        """
        return self.model.value(obs)

    def value_learn(self, obs, val):
        """ Learn the value model with square error cost
        """
        predict_val = self.model.value(obs)
        loss = layers.square_error_cost(predict_val, val)
        loss = layers.reduce_mean(loss)
        optimizer = fluid.optimizer.AdamOptimizer(self.value_lr)
        optimizer.minimize(loss)
        return loss

    def sync_old_policy(self):
        """ Synchronize weights of self.model.policy_model to self.old_policy_model
        """
        self.model.policy_model.sync_weights_to(self.old_policy_model)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = "2.0.1"
"""
generates new PARL python API
"""
import os

from tensorboardX import SummaryWriter
from parl.utils.utils import _HAS_FLUID, _HAS_TORCH, _HAS_PADDLE
from parl.utils import logger

if 'PARL_BACKEND' in os.environ and os.environ['PARL_BACKEND'] != '':
    assert os.environ['PARL_BACKEND'] in ['fluid', 'paddle', 'torch']
    logger.info(
        'Have found environment variable `PARL_BACKEND`==\'{}\', switching backend framework to [{}]'
        .format(os.environ['PARL_BACKEND'], os.environ['PARL_BACKEND']))
    if os.environ['PARL_BACKEND'] == 'paddle':
        from parl.core.paddle import *
    elif os.environ['PARL_BACKEND'] == 'fluid':
        from parl.core.fluid import *
        from parl.core.fluid.plutils.compiler import compile
    elif os.environ['PARL_BACKEND'] == 'torch':
        assert _HAS_TORCH, 'Torch-based PARL requires torch, which is not installed.'
        from parl.core.torch import *
else:
    if _HAS_PADDLE:
        from parl.core.paddle import *
    elif _HAS_FLUID:
        from parl.core.fluid import *
        from parl.core.fluid.plutils.compiler import compile
    elif _HAS_TORCH:
        from parl.core.torch import *

from parl.remote import remote_class, connect
from parl import algorithms



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn.functional as F
from copy import deepcopy
import parl

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

__all__ = ['DDPG']


class DDPG(parl.Algorithm):
    def __init__(self,
                 model,
                 gamma=None,
                 tau=None,
                 actor_lr=None,
                 critic_lr=None):
        assert isinstance(gamma, float)
        assert isinstance(tau, float)
        assert isinstance(actor_lr, float)
        assert isinstance(critic_lr, float)
        self.gamma = gamma
        self.tau = tau
        self.actor_lr = actor_lr
        self.critic_lr = critic_lr

        self.model = model.to(device)
        self.target_model = deepcopy(self.model)
        self.actor_optimizer = torch.optim.Adam(
            self.model.get_actor_params(), lr=actor_lr)
        self.critic_optimizer = torch.optim.Adam(
            self.model.get_critic_params(), lr=critic_lr)

    def predict(self, obs):
        return self.model.policy(obs)

    def learn(self, obs, action, reward, next_obs, terminal):
        critic_loss = self._critic_learn(obs, action, reward, next_obs,
                                         terminal)
        actor_loss = self._actor_learn(obs)

        self.sync_target()
        return critic_loss, actor_loss

    def _critic_learn(self, obs, action, reward, next_obs, terminal):
        # Compute the target Q value
        target_Q = self.target_model.value(next_obs,
                                           self.target_model.policy(next_obs))
        target_Q = reward + ((1. - terminal) * self.gamma * target_Q).detach()

        # Get current Q estimate
        current_Q = self.model.value(obs, action)

        # Compute critic loss
        critic_loss = F.mse_loss(current_Q, target_Q)

        # Optimize the critic
        self.critic_optimizer.zero_grad()
        critic_loss.backward()
        self.critic_optimizer.step()
        return critic_loss

    def _actor_learn(self, obs):
        # Compute actor loss and Update the frozen target models
        actor_loss = -self.model.value(obs, self.model.policy(obs)).mean()

        # Optimize the actor
        self.actor_optimizer.zero_grad()
        actor_loss.backward()
        self.actor_optimizer.step()
        return actor_loss

    def sync_target(self, decay=None):
        if decay is None:
            decay = 1.0 - self.tau
        for param, target_param in zip(self.model.parameters(),
                                       self.target_model.parameters()):
            target_param.data.copy_((1 - decay) * param.data +
                                    decay * target_param.data)



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from parl.core.model_base import ModelBase


class AlgorithmBase(object):
    """`AlgorithmBase` is the base class of the `parl.Algorithm` in different
    frameworks.

    This base class mainly do the following things:
        1. Implements APIs to set or get weights of all `ModelBase` in self.__dict__;
        2. Defines common APIs that `parl.Algorithm` should implement in different frameworks.
    """

    def __init__(self):
        pass

    def get_weights(self):
        """Get weights of all `ModelBase` in self.__dict__.
        
        Note:
            `ModelBase` in list, tuple and dict will be included. But `ModelBase` in
            nested list, tuple and dict won't be included.

        Returns:
            Dict of weights ({attribute name: numpy array/List/Dict})
        """
        model_weights = {}
        for key in self.__dict__.keys():
            value = getattr(self, key)
            if isinstance(value, ModelBase):
                model_weights[key] = value.get_weights()
            elif isinstance(value, list) or isinstance(value, tuple):
                weights_list = []
                for x in value:
                    if isinstance(x, ModelBase):
                        weights_list.append(x.get_weights())
                if weights_list:
                    model_weights[key] = weights_list
            elif isinstance(value, dict):
                weights_dict = {}
                for sub_k, sub_v in value.items():
                    if isinstance(sub_v, ModelBase):
                        weights_dict[sub_k] = sub_v.get_weights()
                if weights_dict:
                    model_weights[key] = weights_dict
        return model_weights

    def set_weights(self, weights):
        """Set weights of all `ModelBase` in self.__dict__.

        Note:
            `ModelBase` in list, tuple and dict will be included. But `ModelBase` in
            nested list, tuple and dict won't be included.

        Args:
            weights (Dict): Dict of weights ({attribute name: numpy array/List/Dict})
        
        """
        for key in self.__dict__.keys():
            value = getattr(self, key)
            if isinstance(value, ModelBase):
                assert key in weights, "weights is inconsistent with current algorithm."
                value.set_weights(weights[key])
            elif isinstance(value, list) or isinstance(value, tuple):
                model_list = []
                for x in value:
                    if isinstance(x, ModelBase):
                        model_list.append(x)
                if model_list:
                    assert key in weights and len(model_list) == len(weights[key]), \
                        "weights is inconsistent with current algorithm."
                    for i, model in enumerate(model_list):
                        model.set_weights(weights[key][i])
            elif isinstance(value, dict):
                model_dict = {}
                for sub_k, sub_v in value.items():
                    if isinstance(sub_v, ModelBase):
                        model_dict[sub_k] = sub_v
                if model_dict:
                    assert key in weights and set(model_dict.keys()) == set(weights[key].keys()), \
                        "weights is inconsistent with current algorithm."
                    for sub_k, model in model_dict.items():
                        model.set_weights(weights[key][sub_k])

    def learn(self, *args, **kwargs):
        """ define learning process, such as how to optimize the model.
        """
        raise NotImplementedError

    def predict(self, *args, **kwargs):
        """ define predicting process, such as using policy model to predict actions when given observations.
        """
        raise NotImplementedError

    def sample(self, *args, **kwargs):
        """ define sampling process, such as using policy model to sample actions when given observations.
        """
        raise NotImplementedError



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

__all__ = ['global_model_helper']


class ModelHelper(object):
    """Model id helper.

    This helper is used to help `parl.Model` generate a new model id
    or register a given model id in a thread-safe way.
    """

    def __init__(self):
        self._registered_ids = set([])
        self.index = 0
        self.lock = threading.Lock()

    def generate_model_id(self):
        """Generate a unique model_id in a thread-safe way.

        Returns:
            String of model id.
        """
        self.lock.acquire()
        model_id = 'parl_model_{}'.format(self.index)
        while model_id in self._registered_ids:
            self.index += 1
            model_id = 'parl_model_{}'.format(self.index)
        self._registered_ids.add(model_id)
        self.index += 1
        self.lock.release()
        return model_id

    def register_model_id(self, model_id):
        """Register given model id in a thread-safe way.
        
        Raises:
            AssertionError: if the model id is already used.
        """
        model_id_used = False
        self.lock.acquire()
        if model_id in self._registered_ids:
            model_id_used = True
        else:
            self._registered_ids.add(model_id)
        self.lock.release()

        assert not model_id_used, "model id `{}` has been used before, please try another model_id.".format(
            model_id)


global_model_helper = ModelHelper()



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import inspect
import numpy as np

from parl.utils import logger
from parl.remote.remote_wrapper import RemoteWrapper
from parl.remote.proxy_wrapper import proxy_wrapper_func
from parl.remote.future_mode import proxy_wrapper_nowait_func


def remote_class(*args, **kwargs):
    """A Python decorator that enables a class to run all its functions
    remotely.

    Each instance of the remote class can be seemed as a task submitted
    to the cluster by the global client, which is created automatically
    when we call parl.connect(master_address). After global client
    submits the task, the master node will send an available job address
    to this remote instance. Then the remote object will send local python
    files, class definition and initialization arguments to the related job.

    In this way, we can run distributed applications easily and efficiently.

    .. code-block:: python

        @parl.remote_class
        class Actor(object):
            def __init__(self, x):
                self.x = x

            def step(self):
                self.x += 1
                return self.x

        actor = Actor()
        actor.step()

        # Set maximum memory usage to 300 MB for each object.
        @parl.remote_class(max_memory=300)
        class LimitedActor(object):
           ...

    Args:
        max_memory (float): Maximum memory (MB) can be used by each remote
                            instance, the unit is in MB and default value is
                            none(unlimited).

    Returns:
        A remote wrapper for the remote class.

    Raises:
        Exception: An exception is raised if the client is not created
                   by `parl.connect(master_address)` beforehand.
    """

    def decorator(cls):
        assert inspect.isclass(
            cls), "Only class can be decorated by `parl.remote_class`."

        # we are not going to create a remote actor in job.py
        if 'XPARL' in os.environ and os.environ['XPARL'] == 'True':
            logger.warning(
                "Note: this object will be runnning as a local object")
            return cls

        RemoteWrapper._original = cls
        RemoteWrapper._max_memory = max_memory

        if wait:
            proxy_wrapper = proxy_wrapper_func(RemoteWrapper)
        else:
            # nowait
            proxy_wrapper = proxy_wrapper_nowait_func(RemoteWrapper)

        proxy_wrapper._original = cls
        return proxy_wrapper

    args_names = ['max_memory', 'wait']
    for key in kwargs:
        assert key in args_names, "Argument `{}` is not supported in the `@parl.remote_class`, supported arguments: {}".format(
            key, args_names)

    max_memory = kwargs.get('max_memory')
    wait = kwargs.get('wait', True)
    """
        Users may pass some arguments to the decorator (e.g., parl.remote_class(10)).
        The following code tries to handle this issue.

        The `args` is different in the following two decorating way, and we should return different wrapper.
        @parl.remote_class     -> args: (<class '__main__.Actor'>,) -> we should return decorator(cls)
        @parl.remote_class(10) -> args: (10,)                       -> we should return decorator
    """
    if len(args) == 1 and callable(args[0]):  # args[0]: cls
        # The first element in the `args` is a class, we should return decorator(cls)
        return decorator(args[0])

    # The first element in the `args` is not a class, we should return decorator
    return decorator



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cloudpickle
import multiprocessing
import os
import psutil
import signal
import socket
import subprocess
import sys
import tempfile
import time
import threading
import warnings
import zmq
from datetime import datetime
import parl
from parl.utils import get_ip_address, to_byte, to_str, logger, _IS_WINDOWS, kill_process
from parl.remote import remote_constants
from parl.remote.message import InitializedWorker
from parl.remote.status import WorkerStatus
from parl.remote.zmq_utils import create_server_socket, create_client_socket
from parl.remote.grpc_heartbeat import HeartbeatServerThread, HeartbeatClientThread
from six.moves import queue
from parl.remote.utils import get_version


class Worker(object):
    """Worker provides the cpu computation resources for the cluster.

    A worker node is connected to the master node and will send its
    computation resources information to the master node. When a worker
    node is created, it will start `cpu_num` empty jobs and these jobs'
    ip addresses will be send to the master node. Further, when an old
    job is killed, worker will start a new job and send the new job ip
    address to the master node.

    To start a worker, we use the following xparl command line api:

    .. code-block:: python

        xparl connect --address localhost:1234 --cpu_num 8

    Attributes:
        master_address (str): Master's ip address.
        request_master_socket (zmq.Context.socket): A socket which sends job
                                                    address to the master node.
        reply_job_socket (zmq.Context.socket): A socket which receives
                                               job_address from the job.
        remove_job_socket (zmq.Context.socket): A socket that receives commands to remove the job from jobs immediately.
                                                Used to remove the dead job immediately and allocate a new job
                                                instead of waiting for the heartbeat failure.
        job_buffer (str): A buffer that stores initialized jobs for providing new jobs in a short time.

    Args:
        master_address (str): IP address of the master node.
        cpu_num (int): Number of cpu to be used on the worker.
    """

    def __init__(self, master_address, cpu_num=None, log_server_port=None):
        self.lock = threading.Lock()
        self.ctx = zmq.Context.instance()
        self.master_address = master_address
        self.master_is_alive = True
        self.worker_is_alive = True
        self.worker_status = None  # initialized at `self._create_jobs`
        self._set_cpu_num(cpu_num)
        self.job_buffer = queue.Queue(maxsize=self.cpu_num)
        self._create_sockets()
        self.check_env_consistency()
        # create log server
        self.log_server_proc, self.log_server_address = self._create_log_server(
            port=log_server_port)

        # create a thread that waits commands from the job to kill the job.
        self.remove_job_thread = threading.Thread(
            target=self._reply_remove_job)
        self.remove_job_thread.setDaemon(True)
        self.remove_job_thread.start()

        self._create_jobs()

        # create a thread that initializes jobs and adds them into the job_buffer
        job_thread = threading.Thread(target=self._fill_job_buffer)
        job_thread.setDaemon(True)
        job_thread.start()

        thread = threading.Thread(target=self._update_worker_status_to_master)
        thread.setDaemon(True)
        thread.start()

    def _set_cpu_num(self, cpu_num=None):
        """set useable cpu number for worker"""
        if cpu_num is not None:
            assert isinstance(
                cpu_num, int
            ), "cpu_num should be INT type, please check the input type."
            self.cpu_num = cpu_num
        else:
            self.cpu_num = multiprocessing.cpu_count()

    def check_env_consistency(self):
        '''Verify that the parl & python version as well as some other packages in 'worker' process
            matches that of the 'master' process'''
        self.request_master_socket.send_multipart(
            [remote_constants.CHECK_VERSION_TAG])
        message = self.request_master_socket.recv_multipart()
        tag = message[0]
        if tag == remote_constants.NORMAL_TAG:
            worker_parl_version = parl.__version__
            worker_python_version_major = str(sys.version_info.major)
            worker_python_version_minor = str(sys.version_info.minor)
            assert worker_parl_version == to_str(message[1]) and worker_python_version_major == to_str(message[2])\
                and worker_python_version_minor == to_str(message[3]),\
                '''Version mismatch: the "master" is of version "parl={}, python={}.{}". However, \
                "parl={}, python={}.{}"is provided in your environment.'''.format(
                        to_str(message[1]), to_str(message[2]), to_str(message[3]),
                        worker_parl_version, worker_python_version_major, worker_python_version_minor
                    )
            worker_pyarrow_version = str(get_version('pyarrow'))
            master_pyarrow_version = to_str(message[4])
            if worker_pyarrow_version != master_pyarrow_version:
                if master_pyarrow_version == 'None':
                    error_message = """"pyarrow" is provided in your current enviroment, however, it is not \
found in "master"'s environment. To use "pyarrow" for serialization, please install \
"pyarrow={}" in "master"'s environment!""".format(worker_pyarrow_version)
                elif worker_pyarrow_version == 'None':
                    error_message = """"pyarrow" is provided in "master"'s enviroment, however, it is not \
found in your current environment. To use "pyarrow" for serialization, please install \
"pyarrow={}" in your current environment!""".format(master_pyarrow_version)
                else:
                    error_message = '''Version mismatch: the 'master' is of version 'pyarrow={}'. However, \
'pyarrow={}'is provided in your current environment.'''.format(
                        master_pyarrow_version, worker_pyarrow_version)
                raise Exception(error_message)
        else:
            raise NotImplementedError

    def _create_sockets(self):
        """Each worker maintains four sockets:

        (1) request_master_socket: sends job address to master node.
        (2) reply_job_socket: receives job_address from subprocess.
        (3) remove_job_socket : receives commands to remove the job from jobs immediately.
                                Used to remove the dead job immediately and allocate a new job
                                instead of waiting for the heartbeat failure.
        (4) reply_log_server_socket: receives log_server_heartbeat_address from subprocess.

        When a job starts, a new heartbeat socket is created to receive
        heartbeat signals from the job.

        """
        self.worker_ip = get_ip_address()

        # request_master_socket: sends job address to master
        self.request_master_socket = self.ctx.socket(zmq.REQ)
        self.request_master_socket.linger = 0

        # wait for 0.5 second to check whether master is started
        self.request_master_socket.setsockopt(zmq.RCVTIMEO, 500)
        self.request_master_socket.connect("tcp://" + self.master_address)

        # reply_job_socket: receives job_address from subprocess
        self.reply_job_socket = self.ctx.socket(zmq.REP)
        self.reply_job_socket.linger = 0
        reply_job_port = self.reply_job_socket.bind_to_random_port("tcp://*")
        self.reply_job_address = "{}:{}".format(self.worker_ip, reply_job_port)

        # remove_job_socket
        self.remove_job_socket = self.ctx.socket(zmq.REP)
        self.remove_job_socket.linger = 0
        remove_job_port = self.remove_job_socket.bind_to_random_port("tcp://*")
        self.remove_job_address = "{}:{}".format(self.worker_ip,
                                                 remove_job_port)

        # reply_log_server_socket: receives log_server_heartbeat_address from subprocess
        self.reply_log_server_socket, reply_log_server_port = create_server_socket(
            self.ctx)
        self.reply_log_server_address = "{}:{}".format(self.worker_ip,
                                                       reply_log_server_port)

    def _create_jobs(self):
        """Create jobs and send a instance of ``InitializedWorker`` that contains the worker information to the master."""
        try:
            self.request_master_socket.send_multipart(
                [remote_constants.WORKER_CONNECT_TAG])
            _ = self.request_master_socket.recv_multipart()
        except zmq.error.Again as e:
            logger.error("Can not connect to the master, "
                         "please check if master is started.")
            self.master_is_alive = False
            return

        initialized_jobs = self._init_jobs(job_num=self.cpu_num)
        self.request_master_socket.setsockopt(
            zmq.RCVTIMEO, remote_constants.HEARTBEAT_TIMEOUT_S * 1000)

        def master_heartbeat_exit_callback_func():
            logger.warning(
                "[Worker] lost connection with the master, will exit reply heartbeat for master."
            )
            if self.worker_status is not None:
                self.worker_status.clear()
            self.log_server_proc.kill()
            self.log_server_proc.wait()
            # exit the worker
            self.exit()

        self.master_heartbeat_thread = HeartbeatServerThread(
            heartbeat_exit_callback_func=master_heartbeat_exit_callback_func)
        self.master_heartbeat_thread.setDaemon(True)
        self.master_heartbeat_thread.start()
        self.master_heartbeat_address = self.master_heartbeat_thread.get_address(
        )

        logger.set_dir(
            os.path.expanduser('~/.parl_data/worker/{}'.format(
                self.master_heartbeat_address.replace(':', '_'))))
        logger.info("[Worker] Connect to the master node successfully. "
                    "({} CPUs)".format(self.cpu_num))

        for job in initialized_jobs:
            job.worker_address = self.master_heartbeat_address

        initialized_worker = InitializedWorker(self.master_heartbeat_address,
                                               initialized_jobs, self.cpu_num,
                                               socket.gethostname())
        self.request_master_socket.send_multipart([
            remote_constants.WORKER_INITIALIZED_TAG,
            cloudpickle.dumps(initialized_worker)
        ])

        _ = self.request_master_socket.recv_multipart()
        self.worker_status = WorkerStatus(self.master_heartbeat_address,
                                          initialized_jobs, self.cpu_num)

    def _fill_job_buffer(self):
        """An endless loop that adds initialized job into the job buffer"""
        initialized_jobs = []
        while self.worker_is_alive:
            if self.job_buffer.full() is False:
                job_num = self.cpu_num - self.job_buffer.qsize()
                if job_num > 0:
                    initialized_jobs = self._init_jobs(job_num=job_num)
                    for job in initialized_jobs:
                        self.job_buffer.put(job)

            time.sleep(0.02)
        self.exit()

    def _init_jobs(self, job_num):
        """Create jobs.

        Args:
            job_num(int): the number of jobs to create.
        """
        job_file = __file__.replace('worker.pyc', 'job.py')
        job_file = job_file.replace('worker.py', 'job.py')
        command = [
            sys.executable, job_file, "--worker_address",
            self.reply_job_address, "--log_server_address",
            self.log_server_address
        ]

        if sys.version_info.major == 3:
            warnings.simplefilter("ignore", ResourceWarning)

        # avoid that many jobs are killed and restarted at the same time.
        self.lock.acquire()

        # Redirect the output to DEVNULL
        FNULL = open(os.devnull, 'w')
        for _ in range(job_num):
            subprocess.Popen(command, stdout=FNULL, close_fds=True)
        FNULL.close()

        new_jobs = []
        for _ in range(job_num):
            job_message = self.reply_job_socket.recv_multipart()
            self.reply_job_socket.send_multipart([
                remote_constants.NORMAL_TAG,
                to_byte(self.remove_job_address)
            ])
            initialized_job = cloudpickle.loads(job_message[1])
            new_jobs.append(initialized_job)

            def heartbeat_exit_callback_func(job):
                job.is_alive = False
                logger.warning(
                    "[Worker] lost connection with the job:{}".format(
                        job.job_address))
                if self.master_is_alive and self.worker_is_alive:
                    self._remove_job(job.job_address)

            # a thread for sending heartbeat signals to job
            thread = HeartbeatClientThread(
                initialized_job.worker_heartbeat_address,
                heartbeat_exit_callback_func=heartbeat_exit_callback_func,
                exit_func_args=(initialized_job, ))
            thread.setDaemon(True)
            thread.start()

        self.lock.release()
        assert len(new_jobs) > 0, "init jobs failed"
        return new_jobs

    def _remove_job(self, job_address):
        """Kill a job process and update worker information"""
        success = self.worker_status.remove_job(job_address)
        if success:
            while True:
                initialized_job = self.job_buffer.get()
                initialized_job.worker_address = self.master_heartbeat_address
                if initialized_job.is_alive:
                    self.worker_status.add_job(initialized_job)
                    if not initialized_job.is_alive:  # make sure that the job is still alive.
                        self.worker_status.remove_job(
                            initialized_job.job_address)
                        continue
                else:
                    logger.warning(
                        "[Worker] a dead job found. The job buffer will not accept this one."
                    )
                if initialized_job.is_alive:
                    break

            self.lock.acquire()
            self.request_master_socket.send_multipart([
                remote_constants.NEW_JOB_TAG,
                cloudpickle.dumps(initialized_job),
                to_byte(job_address)
            ])
            _ = self.request_master_socket.recv_multipart()
            self.lock.release()

    def _reply_remove_job(self):
        """Worker starts a thread to wait jobs' commands to remove the job immediately"""
        self.remove_job_socket.linger = 0
        self.remove_job_socket.setsockopt(
            zmq.RCVTIMEO, remote_constants.HEARTBEAT_RCVTIMEO_S * 1000)
        while self.worker_is_alive and self.master_is_alive:
            try:
                message = self.remove_job_socket.recv_multipart()
                tag = message[0]
                assert tag == remote_constants.KILLJOB_TAG
                to_remove_job_address = to_str(message[1])
                self._remove_job(to_remove_job_address)
                self.remove_job_socket.send_multipart(
                    [remote_constants.NORMAL_TAG])
            except zmq.error.Again as e:
                #detect whether `self.worker_is_alive` is True periodically
                pass

    def _get_worker_status(self):
        now = datetime.strftime(datetime.now(), '%H:%M:%S')
        virtual_memory = psutil.virtual_memory()
        total_memory = round(virtual_memory[0] / (1024**3), 2)
        used_memory = round(virtual_memory[3] / (1024**3), 2)
        vacant_memory = round(total_memory - used_memory, 2)
        if _IS_WINDOWS:
            load_average = round(psutil.getloadavg()[0], 2)
        else:
            load_average = round(os.getloadavg()[0], 2)

        update_status = {
            "vacant_memory": vacant_memory,
            "used_memory": used_memory,
            "load_time": now,
            "load_value": load_average
        }
        return update_status

    def _update_worker_status_to_master(self):
        while self.master_is_alive and self.worker_is_alive:
            worker_status = self._get_worker_status()

            self.lock.acquire()
            try:
                self.request_master_socket.send_multipart([
                    remote_constants.WORKER_STATUS_UPDATE_TAG,
                    to_byte(self.master_heartbeat_address),
                    cloudpickle.dumps(worker_status)
                ])
                message = self.request_master_socket.recv_multipart()
            except zmq.error.Again as e:
                self.master_is_alive = False
            finally:
                self.lock.release()

            time.sleep(remote_constants.HEARTBEAT_INTERVAL_S)

    def _create_log_server(self, port):
        log_server_file = __file__.replace('worker.pyc', 'log_server.py')
        log_server_file = log_server_file.replace('worker.py', 'log_server.py')

        if port is None:
            port = "0"  # `0` means using a random port in flask
        command = [
            sys.executable,
            log_server_file,
            "--port",
            str(port),
            "--log_dir",
            "~/.parl_data/job/",
            "--line_num",
            "500",
            "--worker_address",
            self.reply_log_server_address,
        ]

        if sys.version_info.major == 3:
            warnings.simplefilter("ignore", ResourceWarning)

        if _IS_WINDOWS:
            FNULL = tempfile.TemporaryFile()
        else:
            FNULL = open(os.devnull, 'w')
        log_server_proc = subprocess.Popen(
            command, stdout=FNULL, close_fds=True)
        FNULL.close()

        log_server_address = "{}:{}".format(self.worker_ip, port)

        message = self.reply_log_server_socket.recv_multipart()
        log_server_heartbeat_addr = to_str(message[1])
        self.reply_log_server_socket.send_multipart(
            [remote_constants.NORMAL_TAG])

        def heartbeat_exit_callback_func():
            # only output warning
            logger.warning("[Worker] lost connection with the log_server.")

        # a thread for sending heartbeat signals to log_server
        thread = HeartbeatClientThread(
            log_server_heartbeat_addr,
            heartbeat_exit_callback_func=heartbeat_exit_callback_func)
        thread.setDaemon(True)
        thread.start()

        return log_server_proc, log_server_address

    def exit(self):
        """close the worker"""
        self.worker_is_alive = False
        kill_process('remote/job.py.*{}'.format(self.reply_job_address))
        if self.master_heartbeat_thread.is_alive():
            self.master_heartbeat_thread.exit()

    def run(self):
        """Keep running until it lost connection with the master.
        """
        if self.worker_is_alive:
            self.master_heartbeat_thread.join()



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file is used to fix the problem that cloudpickle cannot load some packages normally in Mac OS.
We hack the problem by trying load these packages in the main module in advance.

Template:

try:
    import [PACKAGE1]
except ImportError:
    pass

try:
    import [PACKAGE2]
except ImportError:
    pass

"""
from parl.utils import _IS_MAC

if _IS_MAC:
    try:
        import rlschool
    except ImportError:
        pass



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ModelBase(object):
    """`ModelBase` is the base class of the `parl.Model` in different frameworks.

    This base class mainly do the following things:
        1. Implements APIs to manage model_id of the `parl.Model`; 
        2. Defines common APIs that `parl.Model` should implement in different frameworks.
    """

    def __init__(self):
        pass

    def forward(self, *args, **kwargs):
        """Define forward network of the model.
        """
        raise NotImplementedError

    def get_weights(self, *args, **kwargs):
        """Get weights of the model.
        """
        raise NotImplementedError

    def set_weights(self, weights, *args, **kwargs):
        """Set weights of the model with given weights.
        """
        raise NotImplementedError

    def sync_weights_to(self, other_model):
        """Synchronize weights of the model to another model.
        """
        raise NotImplementedError

    def parameters(self):
        """Get the parameters of the model.
        """
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        """Call forward function.
        """
        return self.forward(*args, **kwargs)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class AgentBase(object):
    """`AgentBase` is the base class of the `parl.Agent` in different frameworks.

    `parl.Agent` is responsible for the general data flow outside the algorithm.
    """

    def __init__(self, algorithm):
        """

        Args:
            algorithm (`AlgorithmBase`): an instance of `AlgorithmBase`
        """
        self.alg = algorithm

    def get_weights(self, *args, **kwargs):
        """Get weights of the agent.
        
        Returns:
            (Dict): Dict of weights ({attribute name: numpy array/List/Dict})
        """
        return self.alg.get_weights(*args, **kwargs)

    def set_weights(self, weights, *args, **kwargs):
        """Set weights of the agent with given weights.

        Args:
            weights (Dict): Dict of weights
        """
        self.alg.set_weights(weights, *args, **kwargs)

    def learn(self, *args, **kwargs):
        """The training interface for Agent.
        
        This function will usually do the following things:
            1. Accept numpy data as input;
            2. Feed numpy data or onvert numpy data to tensor (optional);
            3. Call learn function in `Algorithm`.
        """
        raise NotImplementedError

    def predict(self, *args, **kwargs):
        """Predict the action when given the observation of the enviroment.

        In general, this function is used in test process.

        This function will usually do the following things:
           1. Accept numpy data as input;
           2. Feed numpy data or onvert numpy data to tensor (optional);
           3. Call predict function in `Algorithm`.
        """
        raise NotImplementedError

    def sample(self, *args, **kwargs):
        """Sample the action when given the observation of the enviroment.
            
        In general, this function is used in train process.

        This function will usually do the following things:
           1. Accept numpy data as input;
           2. Feed numpy data or onvert numpy data to tensor (optional);
           3. Call predict or sample function in `Algorithm`;
           4. Add sampling operation in numpy level. (unnecessary if sampling operation have done in `Algorithm`).
        """
        raise NotImplementedError



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import zmq
from parl.remote import remote_constants


def create_server_socket(ctx, heartbeat_timeout=False):
    """Create a server socket with a random port (support raising timeout exception).

    Args:
        ctx(zmq.Context()): context of zmq
        heartbeat_timeout(bool): whether to set the timeout(HEARTBEAT_RCVTIMEO_S) for
               receiving operation on the server socket. (The default value is False)

    Returns:
        socket(zmq.Context().socket): socket of the server.
        port(int): port of the server socket.
    """
    socket = ctx.socket(zmq.REP)
    if heartbeat_timeout:
        socket.setsockopt(zmq.RCVTIMEO,
                          remote_constants.HEARTBEAT_RCVTIMEO_S * 1000)
    socket.linger = 0
    port = socket.bind_to_random_port(addr="tcp://*")
    return socket, port


def create_client_socket(ctx, server_socket_address, heartbeat_timeout=False):
    """Create a client socket to connect the `server_socket_address`
    (support raising timeout exception).

    Args:
        ctx(zmq.Context()): context of zmq
        server_socket_address(str): address of server socket
        heartbeat_timeout(bool): whether to set the timeout(HEARTBEAT_RCVTIMEO_S) for
               sending operation on the client socket. (The default value is False)

    Returns:
        socket(zmq.Context().socket): socket of the client.
    """
    socket = ctx.socket(zmq.REQ)
    if heartbeat_timeout:
        socket.setsockopt(zmq.RCVTIMEO,
                          remote_constants.HEARTBEAT_RCVTIMEO_S * 1000)
    socket.linger = 0
    socket.connect("tcp://{}".format(server_socket_address))

    return socket



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import sys
import os
import re
from setuptools import setup, find_packages

cur_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(cur_dir, 'README.md'), 'rb') as f:
    lines = [x.decode('utf-8') for x in f.readlines()]
    lines = ''.join([re.sub('^<.*>\n$', '', x) for x in lines])
    long_description = lines


def _find_packages(prefix=''):
    packages = []
    path = '.'
    prefix = prefix
    for root, _, files in os.walk(path):
        if '__init__.py' in files:
            if sys.platform == 'win32':
                packages.append(
                    re.sub('^[^A-z0-9_]', '', root.replace('\\', '.')))
            else:
                packages.append(
                    re.sub('^[^A-z0-9_]', '', root.replace('/', '.')))
    return packages


def read(*parts):
    with codecs.open(os.path.join(cur_dir, *parts), 'r') as fp:
        return fp.read()


# Reference: https://github.com/pypa/pip/blob/master/setup.py
def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file,
        re.M,
    )
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")


setup(
    name='parl',
    version=find_version("parl", "__init__.py"),
    description='Reinforcement Learning Framework',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/PaddlePaddle/PARL',
    packages=_find_packages(),
    include_package_data=True,
    package_data={'': ['*.so']},
    install_requires=[
        "termcolor>=1.1.0",
        "pyzmq==18.1.1",
        "scipy>=1.0.0",
        'cloudpickle==1.3.0; python_version<"3"',
        'cloudpickle==1.6.0; python_version>="3"',
        "tensorboardX==1.8",
        "tb-nightly==1.15.0a20190801",
        "flask>=1.0.4",
        "click",
        "psutil>=5.6.2",
        "flask_cors",
        "requests",
        "grpcio>=1.27.2",
        "protobuf>=3.14.0",
        "visualdl>=2.0.0b;python_version>='3.8' and platform_system=='Linux'",
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    entry_points={"console_scripts": ["xparl=parl.remote.scripts:main"]},
)



#   Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import parl
import paddle

__all__ = ['DDQN']


class DDQN(parl.Algorithm):
    def __init__(self, model, gamma=None, lr=None):
        """ DDQN algorithm

        Args:
            model (parl.Model): forward neural network representing the Q function.
            gamma (float): discounted factor for `accumulative` reward computation
            lr (float): learning rate.
        """
        self.model = model
        self.target_model = copy.deepcopy(model)

        assert isinstance(gamma, float)
        assert isinstance(lr, float)

        self.gamma = gamma
        self.lr = lr

        self.mse_loss = paddle.nn.MSELoss(reduction='mean')
        self.optimizer = paddle.optimizer.Adam(
            learning_rate=lr, parameters=self.model.parameters())

    def predict(self, obs):
        """ use self.model (Q function) to predict the action values
        """
        return self.model.value(obs)

    def learn(self, obs, action, reward, next_obs, terminal):
        """ update the Q function (self.model) with DDQN algorithm
        """
        # Q
        pred_values = self.model.value(obs)
        action_dim = pred_values.shape[-1]
        action = paddle.squeeze(action, axis=-1)
        action_onehot = paddle.nn.functional.one_hot(
            action, num_classes=action_dim)
        pred_value = paddle.multiply(pred_values, action_onehot)
        pred_value = paddle.sum(pred_value, axis=1, keepdim=True)

        # Q_{target}
        with paddle.no_grad():

            # select greedy action base on Q: a` = argmax_a Q(x`, a)
            greedy_actions = self.model.value(next_obs).argmax(1)

            # get booststrapped next state value: Q_{target}(x`, a`)
            g_action_oh = paddle.nn.functional.one_hot(
                greedy_actions, num_classes=action_dim)
            max_v = self.target_model.value(next_obs).multiply(g_action_oh)
            max_v = max_v.sum(axis=1, keepdim=True)

            # get target value: y_i = r_i + gamma * Q_{target}(x`, a`)
            target = reward + (1 - terminal) * self.gamma * max_v

        loss = self.mse_loss(pred_value, target)

        # optimize
        self.optimizer.clear_grad()
        loss.backward()
        self.optimizer.step()
        return loss

    def sync_target(self):
        self.model.sync_weights_to(self.target_model)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
warnings.simplefilter('default')

from parl.core.fluid import layers
from copy import deepcopy
from paddle import fluid
from parl.core.fluid.algorithm import Algorithm

__all__ = ['DDPG']


class DDPG(Algorithm):
    def __init__(self,
                 model,
                 gamma=None,
                 tau=None,
                 actor_lr=None,
                 critic_lr=None):
        """  DDPG algorithm
        
        Args:
            model (parl.Model): forward network of actor and critic.
                                The function get_actor_params() of model should be implemented.
            gamma (float): discounted factor for reward computation.
            tau (float): decay coefficient when updating the weights of self.target_model with self.model
            actor_lr (float): learning rate of the actor model
            critic_lr (float): learning rate of the critic model
        """
        assert isinstance(gamma, float)
        assert isinstance(tau, float)
        assert isinstance(actor_lr, float)
        assert isinstance(critic_lr, float)
        self.gamma = gamma
        self.tau = tau
        self.actor_lr = actor_lr
        self.critic_lr = critic_lr

        self.model = model
        self.target_model = deepcopy(model)

    def predict(self, obs):
        """ use actor model of self.model to predict the action
        """
        return self.model.policy(obs)

    def learn(self, obs, action, reward, next_obs, terminal):
        """ update actor and critic model with DDPG algorithm
        """
        actor_cost = self._actor_learn(obs)
        critic_cost = self._critic_learn(obs, action, reward, next_obs,
                                         terminal)
        return actor_cost, critic_cost

    def _actor_learn(self, obs):
        action = self.model.policy(obs)
        Q = self.model.value(obs, action)
        cost = layers.reduce_mean(-1.0 * Q)
        optimizer = fluid.optimizer.AdamOptimizer(self.actor_lr)
        optimizer.minimize(cost, parameter_list=self.model.get_actor_params())
        return cost

    def _critic_learn(self, obs, action, reward, next_obs, terminal):
        next_action = self.target_model.policy(next_obs)
        next_Q = self.target_model.value(next_obs, next_action)

        terminal = layers.cast(terminal, dtype='float32')
        target_Q = reward + (1.0 - terminal) * self.gamma * next_Q
        target_Q.stop_gradient = True

        Q = self.model.value(obs, action)
        cost = layers.square_error_cost(Q, target_Q)
        cost = layers.reduce_mean(cost)
        optimizer = fluid.optimizer.AdamOptimizer(self.critic_lr)
        optimizer.minimize(cost)
        return cost

    def sync_target(self, decay=None, share_vars_parallel_executor=None):
        if decay is None:
            decay = 1.0 - self.tau
        self.model.sync_weights_to(
            self.target_model,
            decay=decay,
            share_vars_parallel_executor=share_vars_parallel_executor)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
warnings.simplefilter('default')

import paddle.fluid as fluid
from parl.core.fluid.algorithm import Algorithm
from parl.core.fluid import layers

__all__ = ['PolicyGradient']


class PolicyGradient(Algorithm):
    def __init__(self, model, lr=None):
        """ Policy Gradient algorithm
        
        Args:
            model (parl.Model): forward network of the policy.
            lr (float): learning rate of the policy model.
        """

        self.model = model
        assert isinstance(lr, float)
        self.lr = lr

    def predict(self, obs):
        """ use policy model self.model to predict the action probability
        """
        return self.model(obs)

    def learn(self, obs, action, reward):
        """ update policy model self.model with policy gradient algorithm
        """
        act_prob = self.model(obs)
        log_prob = layers.cross_entropy(act_prob, action)
        cost = log_prob * reward
        cost = layers.reduce_mean(cost)
        optimizer = fluid.optimizer.Adam(self.lr)
        optimizer.minimize(cost)
        return cost



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
warnings.simplefilter('default')

import copy
import torch
import torch.optim as optim
import torch.nn.functional as F
import parl
import numpy as np

__all__ = ['DQN']


class DQN(parl.Algorithm):
    def __init__(self, model, gamma=None, lr=None):
        """ DQN algorithm

        Args:
            model (parl.Model): model defining forward network of Q function.
            gamma (float): discounted factor for reward computation.
            lr (float): learning rate.
        """
        self.model = model
        self.target_model = copy.deepcopy(model)
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.model.to(device)
        self.target_model.to(device)

        assert isinstance(gamma, float)
        assert isinstance(lr, float)
        self.gamma = gamma
        self.lr = lr

        self.mse_loss = torch.nn.MSELoss()
        self.optimizer = optim.Adam(self.model.parameters(), lr=self.lr)

    def predict(self, obs):
        """ use value model self.model to predict the action value
        """
        with torch.no_grad():
            pred_q = self.model(obs)
        return pred_q

    def learn(self, obs, action, reward, next_obs, terminal):
        """ update value model self.model with DQN algorithm
        """
        pred_value = self.model(obs).gather(1, action)
        with torch.no_grad():
            max_v = self.target_model(next_obs).max(1, keepdim=True)[0]
            target = reward + (1 - terminal) * self.gamma * max_v
        self.optimizer.zero_grad()
        loss = self.mse_loss(pred_value, target)
        loss.backward()
        self.optimizer.step()
        return loss.item()

    def sync_target(self):
        self.model.sync_weights_to(self.target_model)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import paddle.fluid as fluid
from parl.core.fluid.layers.layer_wrappers import LayerFunc
from parl.core.fluid.plutils import *
from parl.core.model_base import ModelBase
from parl.utils import machine_info
from parl.core.fluid.model_helper import global_model_helper

__all__ = ['Model']


class Model(ModelBase):
    """
    | `alias`: ``parl.Model``
    | `alias`: ``parl.core.fluid.agent.Model``

    | ``Model`` is a base class of PARL for the neural network. A ``Model`` is usually a policy or Q-value function, which predicts an action or an estimate according to the environmental observation.

    | To track all the layers , users are required to implement neural networks with the layers from ``parl.layers`` (e.g., parl.layers.fc). These layers has the same APIs as fluid.layers.

    | ``Model`` supports duplicating a ``Model`` instance in a pythonic way:

    | ``copied_model = copy.deepcopy(model)``

    Example:

    .. code-block:: python

        import parl

        class Policy(parl.Model):
            def __init__(self):
                self.fc = parl.layers.fc(size=12, act='softmax')

            def policy(self, obs):
                out = self.fc(obs)
                return out
               
        policy = Policy() 
        copied_policy = copy.deepcopy(model)

    Attributes:
        model_id(str): each model instance has its unique model_id.

    Public Functions:
        - ``sync_weights_to``: synchronize parameters of the current model to another model.
        - ``get_weights``: return a list containing all the parameters of the current model.
        - ``set_weights``: copy parameters from ``set_weights()`` to the model.
        - ``forward``: define the computations of a neural network. **Should** be overridden by all subclasses.
        - ``parameters``: return a list containting names of parameters of the model. 
        - ``set_model_id``: set ``model_id`` of current model explicitly.
        - ``get_model_id``: return the ``model_id`` of current model.

    """

    def __init__(self, model_id=None):
        """
        Args:
            model_id (String): user-specified model_id (default: None)
        """
        super(Model, self).__init__()
        if model_id is not None:
            global_model_helper.register_model_id(model_id)
            self.__model_id = model_id
        else:
            self.__model_id = global_model_helper.generate_model_id()

    @property
    def model_id(self):
        return self.get_model_id()

    @model_id.setter
    def model_id(self, model_id):
        self.set_model_id(model_id)

    def get_model_id(self):
        """Get model_id of `ModelBase`.
        If not created, will create a new model_id.

        Returns:
            String of model_id.
        """
        try:
            return self.__model_id
        except AttributeError:
            self.__model_id = global_model_helper.generate_model_id()
            return self.__model_id

    def set_model_id(self, model_id):
        """Set model_id of `ModelBase` with given model_id.
        
        Args:
            model_id (string): string of model_id.
        """
        global_model_helper.register_model_id(model_id)
        self.__model_id = model_id

    def sync_weights_to(self,
                        target_model,
                        decay=0.0,
                        share_vars_parallel_executor=None):
        """Synchronize parameters of current model to another model.
        
        To speed up the synchronizing process, it will create a program implicitly to finish the process. It
        also stores a program as the cache to avoid creating program repeatedly.

        target_model_weights = decay * target_model_weights + (1 - decay) * current_model_weights

        Args:
            target_model (`parl.Model`): an instance of ``Model`` that has the same neural network architecture as the current model.
            decay (float):  the rate of decline in copying parameters. 0 if no parameters decay when synchronizing the parameters.
            share_vars_parallel_executor (fluid.ParallelExecutor): Optional. If not None, will use ``fluid.ParallelExecutor``
                                                                   to run program instead of ``fluid.Executor``.

        Example:

        .. code-block:: python

            import copy
            # create a model that has the same neural network structures.
            target_model = copy.deepcopy(model)

            # after initilizing the parameters ...
            model.sync_weights_to(target_mdodel)

        Note:
            Before calling ``sync_weights_to``, parameters of the model must have been initialized.
        """

        args_hash_id = hashlib.md5('{}_{}'.format(
            id(target_model), decay).encode('utf-8')).hexdigest()
        has_cached = False
        try:
            if self._cached_id == args_hash_id:
                has_cached = True
        except AttributeError:
            has_cached = False

        if not has_cached:
            # Can not run _cached program, need create a new program
            self._cached_id = args_hash_id

            assert not target_model is self, "cannot copy between identical model"
            assert isinstance(target_model, Model)
            assert self.__class__.__name__ == target_model.__class__.__name__, \
                "must be the same class for params syncing!"
            assert (decay >= 0 and decay <= 1)

            param_pairs = self._get_parameter_pairs(self, target_model)

            self._cached_sync_weights_program = fluid.Program()

            with fluid.program_guard(self._cached_sync_weights_program):
                for (src_var_name, target_var_name) in param_pairs:
                    src_var = fetch_framework_var(src_var_name)
                    target_var = fetch_framework_var(target_var_name)
                    fluid.layers.assign(
                        decay * target_var + (1 - decay) * src_var, target_var)

            if share_vars_parallel_executor is None:
                # use fluid.Executor
                place = fluid.CUDAPlace(0) if machine_info.is_gpu_available(
                ) else fluid.CPUPlace()
                self._cached_fluid_executor = fluid.Executor(place)
            else:
                # use fluid.ParallelExecutor

                # specify strategy to make ParallelExecutor run faster
                exec_strategy = fluid.ExecutionStrategy()
                exec_strategy.use_experimental_executor = True
                exec_strategy.num_threads = 4
                build_strategy = fluid.BuildStrategy()
                build_strategy.remove_unnecessary_lock = True

                with fluid.scope_guard(fluid.global_scope().new_scope()):
                    self._cached_fluid_executor = fluid.ParallelExecutor(
                        use_cuda=machine_info.is_gpu_available(),
                        main_program=self._cached_sync_weights_program,
                        share_vars_from=share_vars_parallel_executor,
                        exec_strategy=exec_strategy,
                        build_strategy=build_strategy,
                    )
        if share_vars_parallel_executor is None:
            self._cached_fluid_executor.run(self._cached_sync_weights_program)
        else:
            self._cached_fluid_executor.run(fetch_list=[])

    def parameters(self):
        """Get names of all parameters in this ``Model``.

        Only parameters created by ``parl.layers`` are included.
        The order of parameter names is consistent among
        different instances of the same `Model`.

        Returns:
            param_names(list): list of string containing parameter names of all parameters

        Example:

        .. code-block:: python

            model = Model()
            model.parameters()

            # output: 
            ['fc0.w0', 'fc0.bias0']
            
        """
        try:
            return self._parameter_names
        except AttributeError:
            self._parameter_names = self._get_parameter_names(self)
            return self._parameter_names

    def get_weights(self):
        """Returns a Python list containing parameters of current model.

        Returns: a Python list containing the parameters of current model.
        """
        weights = []
        for param_name in self.parameters():
            weight = fetch_value(param_name)
            weights.append(weight)

        return weights

    def set_weights(self, weights):
        """Copy parameters from ``set_weights()`` to the model.
        
        Args:
            weights (list): a Python list containing the parameters.
        """
        assert len(weights) == len(self.parameters()), \
                'size of input weights should be same as weights number of current model'
        try:
            is_gpu_available = self._is_gpu_available
        except AttributeError:
            self._is_gpu_available = machine_info.is_gpu_available()
            is_gpu_available = self._is_gpu_available

        for (param_name, weight) in list(zip(self.parameters(), weights)):
            set_value(param_name, weight, is_gpu_available)

    def _get_parameter_names(self, obj):
        """ Recursively get parameter names in an object.

        Args:
            obj (Object): any object

        Returns:
            parameter_names (list): all parameter names in this object.
        """

        parameter_names = []
        if isinstance(obj, Model):
            for attr in sorted(obj.__dict__.keys()):
                val = getattr(obj, attr)
                parameter_names.extend(self._get_parameter_names(val))
        elif isinstance(obj, LayerFunc):
            for attr in obj.attr_holder.sorted():
                if attr:
                    parameter_names.append(attr.name)
        elif isinstance(obj, tuple) or isinstance(obj, list):
            for x in obj:
                parameter_names.extend(self._get_parameter_names(x))
        elif isinstance(obj, dict):
            for x in list(obj.values()):
                parameter_names.extend(self._get_parameter_names(x))
        else:
            # for any other type, won't be handled. E.g. set
            pass
        return parameter_names

    def _get_parameter_pairs(self, src, target):
        """ Recursively gets parameters in source model and 
        corresponding parameters in target model.

        Args:
            src (`Model`/`LayerFunc`/list/tuple/dict): source object
            target (`Model`/`LayerFunc`/list/tuple/dict): target object

        Returns:
            param_pairs (list of tuple): all parameter names in source model
                                         and corresponding parameter names in 
                                         target model.
        """

        param_pairs = []
        if isinstance(src, Model):
            for attr in src.__dict__:
                if not attr in target.__dict__:
                    continue
                src_var = getattr(src, attr)
                target_var = getattr(target, attr)
                param_pairs.extend(
                    self._get_parameter_pairs(src_var, target_var))
        elif isinstance(src, LayerFunc):
            src_attrs = src.attr_holder.sorted()
            target_attrs = target.attr_holder.sorted()
            assert len(src_attrs) == len(target_attrs), \
                    "number of ParamAttr between source layer and target layer should be same."
            for (src_attr, target_attr) in zip(src_attrs, target_attrs):
                if src_attr:
                    assert target_attr, "ParamAttr between source layer and target layer is inconsistent."
                    param_pairs.append((src_attr.name, target_attr.name))
        elif isinstance(src, tuple) or isinstance(src, list):
            for src_var, target_var in zip(src, target):
                param_pairs.extend(
                    self._get_parameter_pairs(src_var, target_var))
        elif isinstance(src, dict):
            for k in src.keys():
                assert k in target
                param_pairs.extend(
                    self._get_parameter_pairs(src[k], target[k]))
        else:
            # for any other type, won't be handled. E.g. set
            pass
        return param_pairs



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cloudpickle
import threading
from collections import defaultdict, deque
from parl.utils import to_str


class ClusterMonitor(object):
    """The client monitor watches the cluster status.

    Attributes:
        status (dict): A dict to store workers status and clients status.
    """

    def __init__(self):
        self.status = {
            'workers': defaultdict(dict),
            'clients': defaultdict(dict),
            'client_jobs': defaultdict(dict),
        }
        self.lock = threading.Lock()

    def add_worker_status(self, worker_address, hostname):
        """Record worker status when it is connected to the cluster.
        
        Args:
            worker_address (str): worker ip address
            hostname (str): worker hostname 
        """
        self.lock.acquire()
        worker_status = self.status['workers'][worker_address]
        worker_status['load_value'] = deque(maxlen=10)
        worker_status['load_time'] = deque(maxlen=10)
        worker_status['hostname'] = hostname
        self.lock.release()

    def add_client_job(self, client_id, job_info):
        self.lock.acquire()
        self.status['client_jobs'][client_id].update(job_info)
        self.lock.release()

    def update_client_status(self, client_address, client_status):
        """Update client status with message send from client heartbeat.
        
        Args:
            client_address (str): client ip address.
            client_status (dict): client status information
                                  (hostname, file_path, actor_num, elapsed_time).
        """
        self.lock.acquire()
        self.status['clients'][client_address] = client_status
        self.lock.release()

    def update_worker_status(self, update_status, worker_address, vacant_cpus,
                             total_cpus):
        """Update a worker status.

        Args:
            update_status (dict): worker updated status information 
                                (vacant_memory, used_memory, load_time, load_value).
            worker_address (str): worker ip address.
            vacant_cpus (int): vacant cpu number.
            total_cpus (int): total cpu number.
        """
        self.lock.acquire()
        worker_status = self.status['workers'][worker_address]
        worker_status['vacant_memory'] = update_status['vacant_memory']
        worker_status['used_memory'] = update_status['used_memory']
        worker_status['load_time'].append(update_status['load_time'])
        worker_status['load_value'].append(update_status['load_value'])

        worker_status['vacant_cpus'] = vacant_cpus
        worker_status['used_cpus'] = total_cpus - vacant_cpus
        self.lock.release()

    def drop_worker_status(self, worker_address):
        """Drop worker status when it exits.

        Args:
            worker_address (str): IP address of the exited worker.
        """
        self.lock.acquire()
        self.status['workers'].pop(worker_address)
        self.lock.release()

    def drop_client_status(self, client_address):
        """Drop client status when it exits.

        Args:
            client_address (str): IP address of the exited client.
        """
        self.lock.acquire()
        if client_address in self.status['clients']:
            self.status['clients'].pop(client_address)
        self.lock.release()

    def get_status_info(self):
        """Return a message of current cluster status."""
        self.lock.acquire()
        worker_num = len(self.status['workers'])
        clients_num = len(self.status['clients'])
        used_cpus = 0
        vacant_cpus = 0
        for worker in self.status['workers'].values():
            used_cpus += worker['used_cpus']
            vacant_cpus += worker['vacant_cpus']
        self.lock.release()
        status_info = "has {} used cpus, {} vacant cpus.".format(
            used_cpus, vacant_cpus)
        return status_info

    def get_status(self):
        """Return a cloudpickled status."""
        self.lock.acquire()
        status = cloudpickle.dumps(self.status)
        self.lock.release()
        return status



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: heartbeat.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='heartbeat.proto',
    package='',
    syntax='proto3',
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=
    b'\n\x0fheartbeat.proto\"\x16\n\x07Request\x12\x0b\n\x03tag\x18\x01 \x01(\x0c\"+\n\x05Reply\x12\x0b\n\x03tag\x18\x01 \x01(\x0c\x12\x15\n\rextra_message\x18\x02 \x01(\t2+\n\rGrpcHeartbeat\x12\x1a\n\x04Send\x12\x08.Request\x1a\x06.Reply\"\x00\x62\x06proto3'
)

_REQUEST = _descriptor.Descriptor(
    name='Request',
    full_name='Request',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='tag',
            full_name='Request.tag',
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=19,
    serialized_end=41,
)

_REPLY = _descriptor.Descriptor(
    name='Reply',
    full_name='Reply',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='tag',
            full_name='Reply.tag',
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='extra_message',
            full_name='Reply.extra_message',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=43,
    serialized_end=86,
)

DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Reply'] = _REPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Request = _reflection.GeneratedProtocolMessageType(
    'Request',
    (_message.Message, ),
    {
        'DESCRIPTOR': _REQUEST,
        '__module__': 'heartbeat_pb2'
        # @@protoc_insertion_point(class_scope:Request)
    })
_sym_db.RegisterMessage(Request)

Reply = _reflection.GeneratedProtocolMessageType(
    'Reply',
    (_message.Message, ),
    {
        'DESCRIPTOR': _REPLY,
        '__module__': 'heartbeat_pb2'
        # @@protoc_insertion_point(class_scope:Reply)
    })
_sym_db.RegisterMessage(Reply)

_GRPCHEARTBEAT = _descriptor.ServiceDescriptor(
    name='GrpcHeartbeat',
    full_name='GrpcHeartbeat',
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=88,
    serialized_end=131,
    methods=[
        _descriptor.MethodDescriptor(
            name='Send',
            full_name='GrpcHeartbeat.Send',
            index=0,
            containing_service=None,
            input_type=_REQUEST,
            output_type=_REPLY,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ])
_sym_db.RegisterServiceDescriptor(_GRPCHEARTBEAT)

DESCRIPTOR.services_by_name['GrpcHeartbeat'] = _GRPCHEARTBEAT

# @@protoc_insertion_point(module_scope)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cloudpickle
import inspect
import sys
import threading
import zmq

from parl.utils import logger, to_str, to_byte
from parl.remote.communication import loads_argument, loads_return,\
    dumps_argument, dumps_return
from parl.remote.client import get_global_client
from parl.remote import remote_constants
from parl.remote.utils import locate_remote_file
from parl.remote.exceptions import RemoteError, RemoteAttributeError,\
    RemoteDeserializeError, RemoteSerializeError, ResourceError, FutureFunctionError
from parl.remote.future_mode.actor_ref_monitor import ActorRefMonitor

XPARL_RESERVED_PREFIX = "_xparl"


class RemoteWrapper(object):
    """
    Wrapper for remote class in client side.
    1. request cpu resource and submit job to the master
    2. return the result of the function called by the user, using the remote computation resource.
    """

    def __init__(self, *args, **kwargs):
        """
        Args:
            args, kwargs: arguments for the initialization of the unwrapped
            class.
        """
        self.GLOBAL_CLIENT = get_global_client()
        self.remote_attribute_keys_set = set()
        self.ctx = self.GLOBAL_CLIENT.ctx

        # GLOBAL_CLIENT will set `master_is_alive` to False when hearbeat
        # finds the master is dead.

        # instance of actor class which is decorated by @remote_class(wait=False).
        # use the reference count of the object to detect whether
        # the object has been deleted or out of scope.
        proxy_wrapper_nowait_object = kwargs.get('_xparl_proxy_wrapper_nowait')
        if proxy_wrapper_nowait_object is None:
            actor_ref_monitor = None
        else:
            actor_ref_monitor = ActorRefMonitor(proxy_wrapper_nowait_object)

        # class which is decorated by @remote_class
        cls = kwargs.get('_xparl_remote_class')

        # max_memory argument in @remote_class decorator
        max_memory = kwargs.get('_xparl_remote_class_max_memory')

        if self.GLOBAL_CLIENT.master_is_alive:
            job_address = self.request_cpu_resource(
                self.GLOBAL_CLIENT, max_memory, actor_ref_monitor)
        else:
            raise Exception("Can not submit job to the master. "
                            "Please check if master is still alive.")

        if job_address is None:
            raise ResourceError("Cannot submit the job to the master. "
                                "Please add more CPU resources to the "
                                "master or try again later.")

        self.internal_lock = threading.Lock()

        # Send actor commands like `init` and `call` to the job.
        self.job_socket = self.ctx.socket(zmq.REQ)
        self.job_socket.linger = 0
        self.job_socket.connect("tcp://{}".format(job_address))
        self.job_address = job_address
        self.job_shutdown = False

        self.send_file(self.job_socket)
        module_path = inspect.getfile(cls)
        if module_path.endswith('pyc'):
            module_path = module_path[:-4]
        elif module_path.endswith('py'):
            module_path = module_path[:-3]
        else:
            raise FileNotFoundError(
                "cannot not find the module:{}".format(module_path))

        if ".." in module_path:
            # append relative path (E.g. "../a/") to the sys.path,
            # inspect.getfile may return an abnormal path (E.g. "/home/user/../a/").
            module_path = module_path[module_path.index(".."):]

        res = inspect.getfile(cls)
        file_path, in_sys_path = locate_remote_file(module_path)
        cls_source = inspect.getsourcelines(cls)
        end_of_file = cls_source[1] + len(cls_source[0])
        class_name = cls.__name__

        for key in list(kwargs.keys()):
            if key.startswith(XPARL_RESERVED_PREFIX):
                del kwargs[key]
        self.job_socket.send_multipart([
            remote_constants.INIT_OBJECT_TAG,
            cloudpickle.dumps(
                [file_path, class_name, end_of_file, in_sys_path]),
            cloudpickle.dumps([args, kwargs]),
            cloudpickle.dumps(sys.path)
        ])
        message = self.job_socket.recv_multipart()
        tag = message[0]
        if tag == remote_constants.NORMAL_TAG:
            self.remote_attribute_keys_set = loads_return(message[1])
        elif tag == remote_constants.EXCEPTION_TAG:
            traceback_str = to_str(message[1])
            self.job_shutdown = True
            raise RemoteError('__init__', traceback_str)
        else:
            pass

    def __del__(self):
        """Delete the remote class object and release remote resources."""
        try:
            if not self.job_shutdown:
                self.job_socket.setsockopt(zmq.RCVTIMEO, 1 * 1000)

                self.job_socket.send_multipart([remote_constants.KILLJOB_TAG])
                _ = self.job_socket.recv_multipart()
                self.job_socket.close(0)

        except AttributeError:
            pass
        except zmq.error.ZMQError:
            pass
        except TypeError:
            pass

    def has_attr(self, attr):
        has_attr = attr in self.remote_attribute_keys_set
        return has_attr

    def get_attrs(self):
        return self.remote_attribute_keys_set

    def send_file(self, socket):
        try:
            socket.send_multipart(
                [remote_constants.SEND_FILE_TAG, self.GLOBAL_CLIENT.pyfiles])
            _ = socket.recv_multipart()
        except zmq.error.Again as e:
            logger.error("Send python files failed.")

    def request_cpu_resource(self, global_client, max_memory,
                             actor_ref_monitor):
        """Try to request cpu resource for 1 second/time for 300 times."""
        cnt = 300
        while cnt > 0:
            job_address = global_client.submit_job(max_memory,
                                                   actor_ref_monitor)
            if job_address is not None:
                return job_address
            if cnt % 30 == 0:
                logger.warning("No vacant cpu resources at the moment, "
                               "will try {} times later.".format(cnt))
            cnt -= 1
        return None

    def set_remote_attr(self, attr, value):
        self.internal_lock.acquire()
        self.job_socket.send_multipart([
            remote_constants.SET_ATTRIBUTE_TAG,
            to_byte(attr),
            dumps_return(value)
        ])
        message = self.job_socket.recv_multipart()
        tag = message[0]
        if tag == remote_constants.NORMAL_TAG:
            self.remote_attribute_keys_set = loads_return(message[1])
            self.internal_lock.release()
        else:
            self.job_shutdown = True
            raise NotImplementedError()
        return

    def get_remote_attr(self, attr):
        """Call the function of the unwrapped class."""
        #check if attr is a attribute or a function
        is_attribute = attr in self.remote_attribute_keys_set

        def wrapper(*args, **kwargs):
            if self.job_shutdown:
                raise RemoteError(attr,
                                  "This actor losts connection with the job.")
            self.internal_lock.acquire()
            if is_attribute:
                self.job_socket.send_multipart(
                    [remote_constants.GET_ATTRIBUTE_TAG,
                     to_byte(attr)])
            else:
                data = dumps_argument(*args, **kwargs)
                self.job_socket.send_multipart(
                    [remote_constants.CALL_TAG,
                     to_byte(attr), data])

            message = self.job_socket.recv_multipart()
            tag = message[0]

            if tag == remote_constants.NORMAL_TAG:
                ret = loads_return(message[1])
                if not is_attribute:
                    self.remote_attribute_keys_set = loads_return(message[2])
                self.internal_lock.release()
                return ret

            elif tag == remote_constants.EXCEPTION_TAG:
                error_str = to_str(message[1])
                self.job_shutdown = True
                raise RemoteError(attr, error_str)

            elif tag == remote_constants.ATTRIBUTE_EXCEPTION_TAG:
                error_str = to_str(message[1])
                self.job_shutdown = True
                raise RemoteAttributeError(attr, error_str)

            elif tag == remote_constants.SERIALIZE_EXCEPTION_TAG:
                error_str = to_str(message[1])
                self.job_shutdown = True
                raise RemoteSerializeError(attr, error_str)

            elif tag == remote_constants.DESERIALIZE_EXCEPTION_TAG:
                error_str = to_str(message[1])
                self.job_shutdown = True
                raise RemoteDeserializeError(attr, error_str)

            else:
                self.job_shutdown = True
                raise NotImplementedError()

        return wrapper() if is_attribute else wrapper



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import os
import time
import threading
from concurrent import futures
from parl.remote import remote_constants
from parl.remote.grpc_heartbeat import heartbeat_pb2
from parl.remote.grpc_heartbeat import heartbeat_pb2_grpc
from parl.utils import logger, get_ip_address


class GrpcHeartbeatServer(heartbeat_pb2_grpc.GrpcHeartbeatServicer):
    def __init__(self):
        self.last_heartbeat_time = time.time()
        self.stop_tag = None
        self.stop_message = None
        self.has_asked_client_to_stop = False
        self.exit_flag = False

    def Send(self, request, context):
        if self.stop_tag is not None:
            self.has_asked_client_to_stop = True
            return heartbeat_pb2.Reply(
                tag=self.stop_tag, extra_message=self.stop_message)

        self.last_heartbeat_time = time.time()
        return heartbeat_pb2.Reply(tag=remote_constants.HEARTBEAT_TAG)

    def stop(self, stop_tag, stop_message):
        """stop the heartbeat server and send the stop_message to the client.
        
        Args:
            stop_tag(byte): tag to inform why stop the heartbeat.
            stop_message(str): error message which will be sent to the client.
        """
        self.stop_tag = stop_tag
        self.stop_message = stop_message

    def exit(self):
        """exit the heartbeat server.
        """
        self.exit_flag = True

    def timeout_timer(self):
        while True:
            time.sleep(remote_constants.HEARTBEAT_INTERVAL_S)

            if self.exit_flag:
                break

            if self.stop_tag is not None and self.has_asked_client_to_stop == True:
                break

            if time.time(
            ) - self.last_heartbeat_time > remote_constants.HEARTBEAT_RCVTIMEO_S:
                # heartbeat exit
                break


class HeartbeatServerThread(threading.Thread):
    def __init__(self,
                 heartbeat_exit_callback_func,
                 exit_func_args=(),
                 exit_func_kwargs={}):
        """Create a thread to run the heartbeat server.

            Args:
                heartbeat_exit_callback_func(function): A callback function, which will be called after the 
                                                        heartbeat exit.
                exit_func_args(tuple): the argument tuple for calling the heartbeat_exit_callback_func. Defaults to ().
                exit_func_kwargs(dict): the argument dict for calling the heartbeat_exit_callback_func. Defaults to {}.
        """
        assert callable(
            heartbeat_exit_callback_func), "It should be a function."
        assert isinstance(exit_func_args, tuple)
        assert isinstance(exit_func_kwargs, dict)

        threading.Thread.__init__(self)
        self.grpc_server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=1),
            options=[('grpc.max_receive_message_length', -1),
                     ('grpc.max_send_message_length', -1)])
        self.heartbeat_server = GrpcHeartbeatServer()

        heartbeat_pb2_grpc.add_GrpcHeartbeatServicer_to_server(
            self.heartbeat_server, self.grpc_server)

        port = self.grpc_server.add_insecure_port('[::]:0')

        self.address = "{}:{}".format(get_ip_address(), port)

        self.heartbeat_exit_callback_func = heartbeat_exit_callback_func
        self._exit_func_args = exit_func_args
        self._exit_func_kwargs = exit_func_kwargs

    def get_address(self):
        return self.address

    def stop(self, stop_tag, stop_message):
        """stop the heartbeat server and send the stop_message to the client.
        
        Args:
            stop_tag(byte): tag to inform why stop the heartbeat.
            stop_message(str): error message which will be sent to the client.
        """
        assert stop_tag in [remote_constants.HEARTBEAT_OUT_OF_MEMORY_TAG], \
                "the stop tag `{}` is not supported".format(stop_tag)

        self.heartbeat_server.stop(stop_tag, stop_message)

    def run(self):
        # unset http_proxy and https_proxy
        if 'http_proxy' in os.environ:
            del os.environ['http_proxy']
        if 'https_proxy' in os.environ:
            del os.environ['https_proxy']

        self.grpc_server.start()

        # a life-long while loop
        self.heartbeat_server.timeout_timer()

        # The heartbeat is exit, try to stop the grpc server.
        self.grpc_server.stop(0)

        # heartbeat is exit, call the exit function.
        self.heartbeat_exit_callback_func(*self._exit_func_args,
                                          **self._exit_func_kwargs)

    def exit(self):
        self.heartbeat_server.exit()



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cloudpickle
import datetime
import os
import socket
import sys
import threading
import zmq
import parl
import time
import glob

from parl.utils import to_str, to_byte, get_ip_address, logger, isnotebook
from parl.remote.utils import get_subfiles_recursively
from parl.remote import remote_constants
from parl.remote.grpc_heartbeat import HeartbeatServerThread, HeartbeatClientThread
from parl.remote.utils import get_version


class Client(object):
    """Base class for the remote client.

    For each training task, there is a global client in the cluster which
    submits jobs to the master node. Different `@parl.remote_class` objects
    connect to the same global client in a training task.

    Attributes:
        submit_job_socket (zmq.Context.socket): A socket which submits job to
                                                the master node.
        pyfiles (bytes): A serialized dictionary containing the code of python
                         files in local working directory.
        executable_path (str): File path of the executable python script.
        start_time (time): A timestamp to record the start time of the program.

    """

    def __init__(self, master_address, process_id, distributed_files=[]):
        """
        Args:
            master_addr (str): ip address of the master node.
            process_id (str): id of the process that created the Client. 
                              Should use os.getpid() to get the process id.
            distributed_files (list): A list of files to be distributed at all
                                      remote instances(e,g. the configuration
                                      file for initialization) .
        """
        self.master_address = master_address
        self.process_id = process_id
        self.ctx = zmq.Context()
        self.lock = threading.Lock()
        self.master_is_alive = True
        self.client_is_alive = True
        self.log_monitor_url = None

        self.executable_path = self.get_executable_path()
        self.all_job_heartbeat_threads = []

        self.actor_num = 0

        self._create_sockets(master_address)
        self.check_env_consistency()

        thread = threading.Thread(target=self._update_client_status_to_master)
        thread.setDaemon(True)
        thread.start()

        self.pyfiles = self.read_local_files(distributed_files)

    def get_executable_path(self):
        """Return current executable path."""
        mod = sys.modules['__main__']
        if hasattr(mod, '__file__'):
            executable_path = os.path.abspath(mod.__file__)
        else:
            executable_path = os.getcwd()
        executable_path = executable_path[:executable_path.rfind('/')]
        return executable_path

    def read_local_files(self, distributed_files=[]):
        """Read local python code and store them in a dictionary, which will
        then be sent to the job.

        Args:
            distributed_files (list): A list of files to be distributed at all
                                      remote instances(e,g. the configuration
                                      file for initialization) . RegExp of file
                                      names is supported. 
                                      e.g. 
                                          distributed_files = ['./*.npy', './test*']
                                                                             
        Returns:
            A cloudpickled dictionary containing the python code in current
            working directory.
        """
        pyfiles = dict()
        pyfiles['python_files'] = {}
        pyfiles['other_files'] = {}

        user_files = []
        user_empty_subfolders = []

        for distributed_file in distributed_files:
            parsed_list = glob.glob(distributed_file)
            if not parsed_list:
                raise ValueError(
                    "no local file is matched with '{}', please check your input"
                    .format(distributed_file))
            for pathname in parsed_list:
                if os.path.isdir(pathname):
                    pythonfiles, otherfiles, emptysubfolders = get_subfiles_recursively(
                        pathname)
                    user_files.extend(pythonfiles)
                    user_files.extend(otherfiles)
                    user_empty_subfolders.extend(emptysubfolders)
                else:
                    user_files.append(pathname)

        if isnotebook():
            main_folder = './'
        else:
            main_file = sys.argv[0]
            main_folder = './'
            sep = os.sep
            if sep in main_file:
                main_folder = sep.join(main_file.split(sep)[:-1])
        code_files = filter(lambda x: x.endswith('.py'),
                            os.listdir(main_folder))

        for file_name in code_files:
            file_path = os.path.join(main_folder, file_name)
            assert os.path.exists(file_path)
            with open(file_path, 'rb') as code_file:
                code = code_file.read()
                pyfiles['python_files'][file_name] = code

        for file_name in set(user_files):
            assert os.path.exists(file_name)
            assert not os.path.isabs(
                file_name
            ), "[XPARL] Please do not distribute a file with absolute path."
            with open(file_name, 'rb') as f:
                content = f.read()
                pyfiles['other_files'][file_name] = content

        pyfiles['empty_subfolders'] = set(user_empty_subfolders)
        return cloudpickle.dumps(pyfiles)

    def _create_sockets(self, master_address):
        """ Each client has 1 sockets as start:

        (1) submit_job_socket: submits jobs to master node.
        """

        # submit_job_socket: submits job to master
        self.submit_job_socket = self.ctx.socket(zmq.REQ)
        self.submit_job_socket.linger = 0
        self.submit_job_socket.setsockopt(
            zmq.RCVTIMEO, remote_constants.HEARTBEAT_TIMEOUT_S * 1000)
        self.submit_job_socket.connect("tcp://{}".format(master_address))
        self.start_time = time.time()

        def master_heartbeat_exit_callback_func():
            logger.warning("[Client] Cannot connect to the master. "
                           "Please check if it is still alive.")
            logger.warning("Client exit replying heartbeat for master.")
            self.master_is_alive = False

        self.master_heartbeat_thread = HeartbeatServerThread(
            heartbeat_exit_callback_func=master_heartbeat_exit_callback_func)
        self.master_heartbeat_thread.setDaemon(True)
        self.master_heartbeat_thread.start()
        self.reply_master_heartbeat_address = self.master_heartbeat_thread.get_address(
        )

        self.client_id = self.reply_master_heartbeat_address.replace(':', '_') + \
                            '_' + str(int(time.time()))

        # check if the master is connected properly
        try:
            self.submit_job_socket.send_multipart([
                remote_constants.CLIENT_CONNECT_TAG,
                to_byte(self.reply_master_heartbeat_address),
                to_byte(socket.gethostname()),
                to_byte(self.client_id),
            ])
            message = self.submit_job_socket.recv_multipart()
            self.log_monitor_url = to_str(message[1])
        except zmq.error.Again as e:
            logger.warning("[Client] Can not connect to the master, please "
                           "check if master is started and ensure the input "
                           "address {} is correct.".format(master_address))
            self.master_is_alive = False
            raise Exception("Client can not connect to the master, please "
                            "check if master is started and ensure the input "
                            "address {} is correct.".format(master_address))

    def check_env_consistency(self):
        '''Verify that the parl & python version as well as some other packages in 'worker' process
            matches that of the 'master' process'''
        self.submit_job_socket.send_multipart(
            [remote_constants.CHECK_VERSION_TAG])
        message = self.submit_job_socket.recv_multipart()
        tag = message[0]
        if tag == remote_constants.NORMAL_TAG:
            client_parl_version = parl.__version__
            client_python_version_major = str(sys.version_info.major)
            client_python_version_minor = str(sys.version_info.minor)
            assert client_parl_version == to_str(message[1]) and client_python_version_major == to_str(message[2])\
                and client_python_version_minor == to_str(message[3]),\
                '''Version mismatch: the 'master' is of version 'parl={}, python={}.{}'. However, \
                'parl={}, python={}.{}'is provided in your environment.'''.format(
                        to_str(message[1]), to_str(message[2]), to_str(message[3]),
                        client_parl_version, client_python_version_major, client_python_version_minor
                    )
            client_pyarrow_version = str(get_version('pyarrow'))
            master_pyarrow_version = to_str(message[4])
            if client_pyarrow_version != master_pyarrow_version:
                if master_pyarrow_version == 'None':
                    error_message = """"pyarrow" is provided in your current enviroment, however, it is not \
found in "master"'s environment. To use "pyarrow" for serialization, please install \
"pyarrow={}" in "master"'s environment!""".format(client_pyarrow_version)
                elif client_pyarrow_version == 'None':
                    error_message = """"pyarrow" is provided in "master"'s enviroment, however, it is not \
found in your current environment. To use "pyarrow" for serialization, please install \
"pyarrow={}" in your current environment!""".format(master_pyarrow_version)
                else:
                    error_message = '''Version mismatch: the 'master' is of version 'pyarrow={}'. However, \
'pyarrow={}'is provided in your current environment.'''.format(
                        master_pyarrow_version, client_pyarrow_version)
                raise Exception(error_message)
        else:
            raise NotImplementedError

    def _update_client_status_to_master(self):
        while self.master_is_alive:
            elapsed_time = datetime.timedelta(
                seconds=int(time.time() - self.start_time))
            client_status = {
                'file_path': self.executable_path,
                'actor_num': self.actor_num,
                'time': str(elapsed_time),
                'log_monitor_url': self.log_monitor_url,
            }

            self.lock.acquire()
            try:
                self.submit_job_socket.send_multipart([
                    remote_constants.CLIENT_STATUS_UPDATE_TAG,
                    to_byte(self.reply_master_heartbeat_address),
                    cloudpickle.dumps(client_status)
                ])
                message = self.submit_job_socket.recv_multipart()
            except zmq.error.Again as e:
                self.master_is_alive = False
            finally:
                self.lock.release()

            time.sleep(remote_constants.HEARTBEAT_INTERVAL_S)

    def _check_and_monitor_job(self, job_heartbeat_address, job_ping_address,
                               max_memory, actor_ref_monitor):
        """ Sometimes the client may receive a job that is dead, thus 
        we have to check if this job is still alive before adding it to the `actor_num`.
        """
        # job_ping_socket: sends ping signal to job
        job_ping_socket = self.ctx.socket(zmq.REQ)
        job_ping_socket.linger = 0
        job_ping_socket.setsockopt(zmq.RCVTIMEO, int(0.9 * 1000))
        job_ping_socket.connect("tcp://" + job_ping_address)
        try:
            job_ping_socket.send_multipart(
                [remote_constants.HEARTBEAT_TAG,
                 to_byte(str(max_memory))])
            job_ping_socket.recv_multipart()
        except zmq.error.Again:
            job_ping_socket.close(0)
            logger.error(
                "[Client] connects to a finished job, will try again, job_ping_address:{}"
                .format(job_ping_address))
            return False
        job_ping_socket.close(0)

        def heartbeat_exit_callback_func():
            self.lock.acquire()
            self.actor_num -= 1
            logger.error(
                '[xparl] lost connection with a job, current actor num: {}'.
                format(self.actor_num))
            self.lock.release()

        # a thread for sending heartbeat signals to job
        job_heartbeat_thread = HeartbeatClientThread(
            job_heartbeat_address,
            heartbeat_exit_callback_func=heartbeat_exit_callback_func)
        self.all_job_heartbeat_threads.append(job_heartbeat_thread)
        job_heartbeat_thread.setDaemon(True)
        job_heartbeat_thread.start()

        if actor_ref_monitor is not None:
            # If `wait` argument is False in `@parl.remote_class` (future mode),
            # the `actor_ref_monitor` is not None. And we need start a thread to
            # detect whether the actor has been deleted according to the reference
            # count of the actor.
            thread = threading.Thread(
                target=self._check_actor_is_alive,
                args=(actor_ref_monitor, job_heartbeat_thread))
            thread.setDaemon(True)
            thread.start()

        return True

    def _check_actor_is_alive(self, actor_ref_monitor, job_heartbeat_thread):
        """A loop to check whether the actor has been deleted.

        Args:
            actor_ref_monitor (ActorRefMonitor): used for detecting whether the actor 
                                                 has been deleted or out of scope;
            job_heartbeat_thread(HeartbeatClientThread): thread used to send the heartbeat signal to the
                                                         job. We will exit the thread after the actor is
                                                         deleted.
        """
        while self.client_is_alive:
            if actor_ref_monitor.is_deleted():
                # terminate the heartbeat of the job and release the cpu resource.
                job_heartbeat_thread.exit()
                break

            time.sleep(5)

    def submit_job(self, max_memory, proxy_wrapper_nowait_object):
        """Send a job to the Master node.

        When a `@parl.remote_class` object is created, the global client
        sends a job to the master node. Then the master node will allocate
        a vacant job from its job pool to the remote object.

        Args:
            max_memory (float): Maximum memory (MB) can be used by each remote
                                instance, the unit is in MB and default value is
                                none(unlimited).
            proxy_wrapper_nowait_object (object): instance of actor class which is decorated by @remote_class(wait=False),
                                use the reference count of the object to detect whether 
                                the object has been deleted or out of scope.

        Returns:
            job_address(str): IP address of the job. None if there is no available CPU in the cluster.
        """
        if self.master_is_alive:

            while True:
                # A lock to prevent multiple actors from submitting job at the same time.
                self.lock.acquire()
                self.submit_job_socket.send_multipart([
                    remote_constants.CLIENT_SUBMIT_TAG,
                    to_byte(self.reply_master_heartbeat_address),
                    to_byte(self.client_id),
                ])
                message = self.submit_job_socket.recv_multipart()
                self.lock.release()

                tag = message[0]

                if tag == remote_constants.NORMAL_TAG:
                    job_address = to_str(message[1])
                    job_heartbeat_address = to_str(message[2])
                    job_ping_address = to_str(message[3])

                    check_result = self._check_and_monitor_job(
                        job_heartbeat_address, job_ping_address, max_memory,
                        proxy_wrapper_nowait_object)
                    if check_result:
                        self.lock.acquire()
                        self.actor_num += 1
                        self.lock.release()
                        return job_address

                # no vacant CPU resources, cannot submit a new job
                elif tag == remote_constants.CPU_TAG:
                    job_address = None
                    # wait 1 second to avoid requesting in a high frequency.
                    time.sleep(1)
                    return job_address
                else:
                    raise NotImplementedError
        else:
            raise Exception("Client can not submit job to the master, "
                            "please check if master is connected.")
        return None


GLOBAL_CLIENT = None


def connect(master_address, distributed_files=[]):
    """Create a global client which connects to the master node.

    .. code-block:: python

        parl.connect(master_address='localhost:1234')

    Args:
        master_address (str): The address of the Master node to connect to.
        distributed_files (list): A list of files to be distributed at all 
                                  remote instances(e,g. the configuration
                                  file for initialization) .

    Raises:
        Exception: An exception is raised if the master node is not started.
    """

    assert len(master_address.split(":")) == 2, "Please input address in " +\
        "{ip}:{port} format"
    assert isinstance(distributed_files,
                      list), "`distributed_files` should be a list."

    global GLOBAL_CLIENT
    addr = master_address.split(":")[0]
    cur_process_id = os.getpid()
    if GLOBAL_CLIENT is None:
        GLOBAL_CLIENT = Client(master_address, cur_process_id,
                               distributed_files)
    else:
        if GLOBAL_CLIENT.process_id != cur_process_id:
            GLOBAL_CLIENT = Client(master_address, cur_process_id,
                                   distributed_files)
    logger.info("Remote actors log url: {}".format(
        GLOBAL_CLIENT.log_monitor_url))


def get_global_client():
    """Get the global client.

    To support process-based programming, we will create a new global client in the new process.

    Returns:
        The global client.
    """
    global GLOBAL_CLIENT
    assert GLOBAL_CLIENT is not None, "Cannot get the client to submit the" +\
        " job, have you connected to the cluster by calling " +\
        "parl.connect(master_ip, master_port)?"

    cur_process_id = os.getpid()
    if GLOBAL_CLIENT.process_id != cur_process_id:
        GLOBAL_CLIENT = Client(GLOBAL_CLIENT.master_address, cur_process_id)
    return GLOBAL_CLIENT


def disconnect():
    """Disconnect the global client from the master node."""
    global GLOBAL_CLIENT
    if GLOBAL_CLIENT is not None:
        GLOBAL_CLIENT.client_is_alive = False
        for thread in GLOBAL_CLIENT.all_job_heartbeat_threads:
            if thread.is_alive():
                thread.exit()

        if GLOBAL_CLIENT.master_heartbeat_thread.is_alive():
            GLOBAL_CLIENT.master_heartbeat_thread.exit()

        GLOBAL_CLIENT = None
    else:
        logger.info(
            "No client to be released. Please make sure that you have called `parl.connect`"
        )



#   Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import parl
import paddle
import paddle.nn.functional as F
from paddle.distribution import Categorical
import numpy as np

__all__ = ['A2C']


class A2C(parl.Algorithm):
    def __init__(self, model, vf_loss_coeff=None):
        """ A2C algorithm

        Args:
            model (parl.Model): forward network of policy and value
            vf_loss_coeff (float): coefficient of the value function loss
        """

        self.model = model
        assert isinstance(vf_loss_coeff, (int, float))
        self.vf_loss_coeff = vf_loss_coeff
        clip = paddle.nn.ClipGradByGlobalNorm(clip_norm=40.0)
        self.optimizer = paddle.optimizer.Adam(
            learning_rate=0.001,
            parameters=self.model.parameters(),
            grad_clip=clip)

    def learn(self, obs, actions, advantages, target_values, learning_rate,
              entropy_coeff):
        """
        Args:
            obs: An float32 tensor of shape ([B] + observation_space).
                 E.g. [B, C, H, W] in atari.
            actions: An int64 tensor of shape [B].
            advantages: A float32 tensor of shape [B].
            target_values: A float32 tensor of shape [B].
            learning_rate: float scalar of leanring rate.
            entropy_coeff: float scalar of entropy coefficient.
        """
        # shape: [B, act_dim]
        logits = self.model.policy(obs)
        act_dim = logits.shape[-1]
        actions_onehot = F.one_hot(actions, act_dim)
        # [B, act_dim] --> [B]
        actions_log_probs = paddle.sum(
            F.log_softmax(logits) * actions_onehot, axis=-1)
        # The policy gradient loss
        pi_loss = -1.0 * paddle.sum(actions_log_probs * advantages)

        # The value function loss
        values = self.model.value(obs)
        delta = values - target_values
        vf_loss = 0.5 * paddle.sum(paddle.square(delta))

        # The entropy loss (We want to maximize entropy, so entropy_ceoff < 0)
        # Using the Categorical just for calculating the entropy.
        # See  https://github.com/PaddlePaddle/Paddle/blob/release/2.0/python/paddle/distribution.py for detail
        policy_distribution = Categorical(logits)
        policy_entropy = policy_distribution.entropy()
        entropy = paddle.sum(policy_entropy)

        total_loss = (
            pi_loss + vf_loss * self.vf_loss_coeff + entropy * entropy_coeff)
        self.optimizer.set_lr(learning_rate)
        total_loss.backward()
        self.optimizer.step()
        self.optimizer.clear_grad()
        return total_loss, pi_loss, vf_loss, entropy

    def prob_and_value(self, obs):
        """
        Args:
            obs: An float32 tensor of shape ([B] + observation_space).
                 E.g. [B, C, H, W] in atari.
        """
        logits, values = self.model.policy_and_value(obs)
        probs = F.softmax(logits)

        return probs, values

    def predict(self, obs):
        """
        Args:
            obs: An float32 tensor of shape ([B] + observation_space).
                 E.g. [B, C, H, W] in atari.
        """
        logits = self.model.policy(obs)
        probs = F.softmax(logits)
        predict_actions = paddle.argmax(probs, 1)
        return predict_actions

    def value(self, obs):
        """
        Args:
            obs: An float32 tensor of shape ([B] + observation_space).
                 E.g. [B, C, H, W] in atari.
        """
        values = self.model.value(obs)
        return values



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from contextlib import contextmanager
import os
import pkg_resources
from parl.utils import isnotebook, logger, format_uniform_path

__all__ = [
    'load_remote_class', 'redirect_output_to_file', 'locate_remote_file',
    'get_subfiles_recursively', 'get_version'
]


def simplify_code(code, end_of_file):
    """
  @parl.remote_actor has to use this function to simplify the code.
  To create a remote object, PARL has to import the module that contains the decorated class.
  It may run some unnecessary code when importing the module, and this is why we use this function
  to simplify the code.

  For example.
  @parl.remote_actor
  class A(object):
    def add(self, a, b):
    return a + b
  def data_process():
    XXXX
  ------------------>
  The last two lines of the above code block will be removed as they are not class-related.
  """
    to_write_lines = []
    for i, line in enumerate(code):
        if line.startswith('parl.connect'):
            continue
        if i < end_of_file - 1:
            to_write_lines.append(line)
        else:
            break
    return to_write_lines


def load_remote_class(file_name, class_name, end_of_file, in_sys_path):
    """
  load a class given its file_name and class_name.

  Args:
    file_name: specify the file to load the class
    class_name: specify the class to be loaded
    end_of_file: line ID to indicate the last line that defines the class.
    in_sys_path: whether the path of the remote class is in the environment path (sys.path).

  Return:
    cls: the class to load
  """
    with open(file_name + '.py') as t_file:
        code = t_file.readlines()
    code = simplify_code(code, end_of_file)
    #folder/xx.py -> folder/xparl_xx.py
    file_name = file_name.split(os.sep)
    prefix = os.sep.join(file_name[:-1])
    if prefix == "":
        prefix = '.'

    # Add pid to fix:
    #    https://github.com/PaddlePaddle/PARL/issues/611
    #    Multiple jobs may write the same file when `in_sys_path` is True.
    new_file_name = 'xparl_{}'.format(os.getpid()) + file_name[-1]
    module_name = prefix + os.sep + new_file_name
    tmp_file_name = module_name + '.py'
    with open(tmp_file_name, 'w') as t_file:
        for line in code:
            t_file.write(line)

    if in_sys_path:
        # the path of the remote class is in the sys.path, we can import it directly.
        mod = __import__(new_file_name)
    else:
        module_name = module_name.lstrip('.' + os.sep).replace(os.sep, '.')
        mod = __import__(module_name, globals(), locals(), [class_name], 0)

    cls = getattr(mod, class_name)

    os.remove(tmp_file_name)

    return cls


@contextmanager
def redirect_output_to_file(stdout_file_path, stderr_file_path):
    """Redirect stdout (e.g., `print`) and stderr (e.g., `warning/error`) to given files respectively.

    Args:
        stdout_file_path: Path of the file to output the stdout.
        stderr_file_path: Path of the file to output the stderr.

    Example:
    >>> print('test')
    test
    >>> with redirect_output_to_file('stdout.log', 'stderr.log'):
    ...     print('test')  # Output nothing, `test` is printed to `stdout.log`.
    >>> print('test')
    test
    """

    origin_stdout = sys.stdout
    origin_stderr = sys.stderr

    stdout_f = open(stdout_file_path, 'a')
    stderr_f = open(stderr_file_path, 'a')

    sys.stdout = stdout_f
    sys.stderr = stderr_f

    # NOTE: we should add the handler after executing the above code.
    handler = logger.add_stdout_handler()

    try:
        yield
    finally:
        sys.stdout = origin_stdout
        sys.stderr = origin_stderr

        stdout_f.close()
        stderr_f.close()

        logger.remove_handler(handler)


def locate_remote_file(module_path):
    """xparl has to locate the file that has the class decorated by parl.remote_class. 

    If the entry_path is the prefix of the absolute path of the module, which means the 
    file of the module is in the directory of entry or its subdirectories, this function
    will return the relative path between this file and the entry file.

    Otherwise, it means that the user accesses the module by adding the path of the
    module to the environment path (sys.path), this function will return the module_path
    directly and the remote job also need access the module by adding the path of the 
    module to the sys.path.

    Note that this function should support the jupyter-notebook environment.

    Args:
        module_path: Absolute path of the module, where the class decorated by
                     parl.remote_class is located.
    
    Returns:
        (return_path, in_sys_path)

        return_path(str): relative path of the module, if the entry_path is the prefix of the absolute
                          path of the module, else module_path.

        in_sys_path(bool): False, if the entry_path is the prefix of the absolute path of the module,
                           else True.
    Example:
        module_path: /home/user/dir/subdir/my_module (or) ./dir/main
        entry_file: /home/user/dir/main.py
        --------> return_path: subdir/my_module

        module_path: /home/user/other_dir/subdir/my_module
        entry_file: /home/user/dir/main.py
        --------> return_path: /home/user/other_dir/subdir/my_module

        module_path: ../other_dir/subdir/my_module
        entry_file: /home/user/dir/main.py
        --------> return_path: ../other_dir/subdir/my_module
    """
    if isnotebook():
        entry_path = os.getcwd()
    else:
        entry_file = sys.argv[0]
        entry_file = entry_file.split(os.sep)[-1]
        entry_path = None
        for path in sys.path:
            to_check_path = os.path.join(path, entry_file)
            if os.path.isfile(to_check_path):
                entry_path = path
                break

    if entry_path is None:
        raise FileNotFoundError("cannot locate the remote file")

    # fix pycharm issue: https://github.com/PaddlePaddle/PARL/issues/350
    module_path = format_uniform_path(module_path)
    entry_path = format_uniform_path(entry_path)

    # transfer the relative path to the absolute path
    abs_module_path = module_path
    if not os.path.isabs(abs_module_path):
        abs_module_path = os.path.abspath(abs_module_path)

    if os.sep in abs_module_path \
            and entry_path != abs_module_path[:len(entry_path)]:
        # the file of the module is not in the directory of entry or its subdirectories
        return module_path, True

    if os.sep in abs_module_path:
        relative_module_path = '.' + abs_module_path[len(entry_path):]
    else:
        relative_module_path = abs_module_path

    return relative_module_path, False


def get_subfiles_recursively(folder_path):
    '''
    Get subfiles under 'folder_path' recursively
    Args:
        folder_path: A folder(dir) whose subfiles/subfolders will be returned.

    Returns:
        python_files: A list including subfiles endwith '.py'.
        other_files: A list including subfiles not endwith '.py'.
        empty_subfolders: A list including empty subfolders.
    '''
    if not os.path.exists(folder_path):
        raise ValueError("Path '{}' don't exist.".format(folder_path))
    elif not os.path.isdir(folder_path):
        raise ValueError('Input should be a folder, not a file.')
    else:
        python_files = []
        other_files = []
        empty_subfolders = []
        for root, dirs, files in os.walk(folder_path):
            if files:
                for sub_file in files:
                    if sub_file.endswith('.py'):
                        python_files.append(
                            os.path.normpath(os.path.join(root, sub_file)))
                    else:
                        other_files.append(
                            os.path.normpath(os.path.join(root, sub_file)))
            elif len(dirs) == 0:
                empty_subfolders.append(os.path.normpath(root))
        return python_files, other_files, empty_subfolders


def get_version(module_name):
    ''' Check if the python environment has installed the module or package.
        return the version of the module if the module is installed,
        return None otherwise.
    Args:
        module_name (str): module to be checked
    Returns:
        has_module: str (if the module is installed) or None
    '''
    assert isinstance(module_name, str), '"module_name" should be a string!'
    try:
        __import__(module_name)
    except ImportError:
        return None
    else:
        module_version = pkg_resources.get_distribution(module_name).version
        return module_version
